import torch
import pytest

from pina import LabelTensor, Condition
from pina.solvers import PINN
from pina.geometry import CartesianDomain
from pina.problem import SpatialProblem
from pina.model import FeedForward
from pina.operators import laplacian
from pina.equation.equation_factory import FixedValue

example_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
example_input_pts = LabelTensor(torch.tensor([[0, 0, 0]]), ['x', 'y', 'z'])
example_output_pts = LabelTensor(torch.tensor([[1, 2]]), ['a', 'b'])


def test_init_inputoutput():
    Condition(input_points=example_input_pts, output_points=example_output_pts)
    with pytest.raises(ValueError):
        Condition(example_input_pts, example_output_pts)
    with pytest.raises(TypeError):
        Condition(input_points=3., output_points='example')
    with pytest.raises(TypeError):
        Condition(input_points=example_domain, output_points=example_domain)


def test_init_locfunc():
    Condition(location=example_domain, equation=FixedValue(0.0))
    with pytest.raises(ValueError):
        Condition(example_domain, FixedValue(0.0))
    with pytest.raises(TypeError):
        Condition(location=3., equation='example')
    with pytest.raises(TypeError):
        Condition(location=example_input_pts, equation=example_output_pts)


def test_init_inputfunc():
    Condition(input_points=example_input_pts, equation=FixedValue(0.0))
    with pytest.raises(ValueError):
        Condition(example_domain, FixedValue(0.0))
    with pytest.raises(TypeError):
        Condition(input_points=3., equation='example')
    with pytest.raises(TypeError):
        Condition(input_points=example_domain, equation=example_output_pts)



"""Module for Intersection class. """

import torch
from ..label_tensor import LabelTensor
from .operation_interface import OperationInterface
import random


class Intersection(OperationInterface):

    def __init__(self, geometries):
        r"""
        PINA implementation of Intersection of Domains.
        Given two sets :math:`A` and :math:`B` then the
        domain difference is defined as:

        .. math::
            A \cap B = \{x \mid x \in A \land x \in B\},

        with :math:`x` a point in :math:`\mathbb{R}^N` and :math:`N`
        the dimension of the geometry space.

        :param list geometries: A list of geometries from ``pina.geometry`` 
            such as ``EllipsoidDomain`` or ``CartesianDomain``. The intersection
            will be taken between all the geometries in the list. The resulting
            geometry will be the intersection of all the geometries in the list.

        :Example:
            >>> # Create two ellipsoid domains
            >>> ellipsoid1 = EllipsoidDomain({'x': [-1, 1], 'y': [-1, 1]})
            >>> ellipsoid2 = EllipsoidDomain({'x': [0, 2], 'y': [0, 2]})
            >>> # Create a Intersection of the ellipsoid domains
            >>> intersection = Intersection([ellipsoid1, ellipsoid2])
        """
        super().__init__(geometries)

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the ``Intersection`` domain.

        :param point: Point to be checked.
        :type point: torch.Tensor   
        :param bool check_border: If ``True``, the border is considered inside.
        :return: ``True`` if the point is inside the Intersection domain, ``False`` otherwise.
        :rtype: bool
        """
        flag = 0
        for geometry in self.geometries:
            if geometry.is_inside(point, check_border):
                flag += 1
        return flag == len(self.geometries)

    def sample(self, n, mode='random', variables='all'):
        """
        Sample routine for ``Intersection`` domain.

        :param int n: Number of points to sample in the shape.
        :param str mode: Mode for sampling, defaults to ``random``. Available modes include: ``random``.
        :param variables: Variables to be sampled, defaults to ``all``.
        :type variables: str | list[str]
        :return: Returns ``LabelTensor`` of n sampled points.
        :rtype: LabelTensor

        :Example:
            >>> # Create two Cartesian domains
            >>> cartesian1 = CartesianDomain({'x': [0, 2], 'y': [0, 2]})
            >>> cartesian2 = CartesianDomain({'x': [1, 3], 'y': [1, 3]})
            >>> # Create a Intersection of the ellipsoid domains
            >>> intersection = Intersection([cartesian1, cartesian2])
            >>> # Sample
            >>> intersection.sample(n=5)
                LabelTensor([[1.7697, 1.8654],
                            [1.2841, 1.1208],
                            [1.7289, 1.9843],
                            [1.3332, 1.2448],
                            [1.9902, 1.4458]])
            >>> len(intersection.sample(n=5)
                5

        """
        if mode != 'random':
            raise NotImplementedError(
                f'{mode} is not a valid mode for sampling.')

        sampled = []

        # calculate the number of points to sample for each geometry and the remainder.
        remainder = n % len(self.geometries)
        num_points = n // len(self.geometries)

        # sample the points
        # NB. geometries as shuffled since if we sample
        # multiple times just one point, we would end
        # up sampling only from the first geometry.
        iter_ = random.sample(self.geometries, len(self.geometries))
        for i, geometry in enumerate(iter_):
            sampled_points = []
            # int(i < remainder) is one only if we have a remainder
            # different than zero. Notice that len(geometries) is
            # always smaller than remaider.
            # makes sure point is uniquely inside 1 shape.
            while len(sampled_points) < (num_points + int(i < remainder)):
                sample = geometry.sample(1, mode, variables)
                if self.is_inside(sample):
                    sampled_points.append(sample)
            sampled += sampled_points

        return LabelTensor(torch.cat(sampled), labels=self.variables)



__all__ = [
    'FeedForward',
    'ResidualFeedForward',
    'MultiFeedForward',
    'DeepONet',
    'MIONet',
    'FNO',
]

from .feed_forward import FeedForward, ResidualFeedForward
from .multi_feed_forward import MultiFeedForward
from .deeponet import DeepONet, MIONet
from .fno import FNO



"""Utils module"""
from torch.utils.data import Dataset, DataLoader
from functools import reduce
import types

import torch
from torch.utils.data import DataLoader, default_collate, ConcatDataset

from .label_tensor import LabelTensor

import torch


def check_consistency(object, object_instance, subclass=False):
    """Helper function to check object inheritance consistency. 
       Given a specific ``'object'`` we check if the object is
       instance of a specific ``'object_instance'``, or in case
       ``'subclass=True'`` we check if the object is subclass
       if the ``'object_instance'``.

    :param (iterable or class object) object: The object to check the inheritance 
    :param Object object_instance: The parent class from where the object
        is expected to inherit
    :param str object_name: The name of the object
    :param bool subclass: Check if is a subclass and not instance
    :raises ValueError: If the object does not inherit from the
        specified class
    """
    if not isinstance(object, (list, set, tuple)):
        object = [object]

    for obj in object:
        try:
            if not subclass:
                assert isinstance(obj, object_instance)
            else:
                assert issubclass(obj, object_instance)
        except AssertionError:
            raise ValueError(f"{type(obj).__name__} must be {object_instance}.")


def number_parameters(model,
                      aggregate=True,
                      only_trainable=True):  # TODO: check
    """
    Return the number of parameters of a given `model`.

    :param torch.nn.Module model: the torch module to inspect.
    :param bool aggregate: if True the return values is an integer corresponding
        to the total amount of parameters of whole model. If False, it returns a
        dictionary whose keys are the names of layers and the values the
        corresponding number of parameters. Default is True.
    :param bool trainable: if True, only trainable parameters are count,
        otherwise no. Default is True.
    :return: the number of parameters of the model
    :rtype: dict or int
    """
    tmp = {}
    for name, parameter in model.named_parameters():
        if only_trainable and not parameter.requires_grad:
            continue

        tmp[name] = parameter.numel()

    if aggregate:
        tmp = sum(tmp.values())

    return tmp


def merge_tensors(tensors):  # name to be changed
    if tensors:
        return reduce(merge_two_tensors, tensors[1:], tensors[0])
    raise ValueError("Expected at least one tensor")


def merge_two_tensors(tensor1, tensor2):
    n1 = tensor1.shape[0]
    n2 = tensor2.shape[0]

    tensor1 = LabelTensor(tensor1.repeat(n2, 1), labels=tensor1.labels)
    tensor2 = LabelTensor(tensor2.repeat_interleave(n1, dim=0),
                          labels=tensor2.labels)
    return tensor1.append(tensor2)


def torch_lhs(n, dim):
    """Latin Hypercube Sampling torch routine.
    Sampling in range $[0, 1)^d$.

    :param int n: number of samples
    :param int dim: dimensions of latin hypercube
    :return: samples
    :rtype: torch.tensor
    """

    if not isinstance(n, int):
        raise TypeError('number of point n must be int')

    if not isinstance(dim, int):
        raise TypeError('dim must be int')

    if dim < 1:
        raise ValueError('dim must be greater than one')

    samples = torch.rand(size=(n, dim))

    perms = torch.tile(torch.arange(1, n + 1), (dim, 1))

    for row in range(dim):
        idx_perm = torch.randperm(perms.shape[-1])
        perms[row, :] = perms[row, idx_perm]

    perms = perms.T

    samples = (perms - samples) / n

    return samples


def is_function(f):
    """
    Checks whether the given object `f` is a function or lambda.

    :param object f: The object to be checked.
    :return: `True` if `f` is a function, `False` otherwise.
    :rtype: bool
    """
    return type(f) == types.FunctionType or type(f) == types.LambdaType


def chebyshev_roots(n):
    """
    Return the roots of *n* Chebyshev polynomials (between [-1, 1]).

    :param int n: number of roots
    :return: roots
    :rtype: torch.tensor
    """
    pi = torch.acos(torch.zeros(1)).item() * 2
    k = torch.arange(n)
    nodes = torch.sort(torch.cos(pi * (k + 0.5) / n))[0]
    return nodes


# class PinaDataset():

#     def __init__(self, pinn) -> None:
#         self.pinn = pinn

#     @property
#     def dataloader(self):
#         return self._create_dataloader()

#     @property
#     def dataset(self):
#         return [self.SampleDataset(key, val)
#                 for key, val in self.input_pts.items()]

#     def _create_dataloader(self):
#         """Private method for creating dataloader

#         :return: dataloader
#         :rtype: torch.utils.data.DataLoader
#         """
#         if self.pinn.batch_size is None:
#             return {key: [{key: val}] for key, val in self.pinn.input_pts.items()}

#         def custom_collate(batch):
#             # extracting pts labels
#             _, pts = list(batch[0].items())[0]
#             labels = pts.labels
#             # calling default torch collate
#             collate_res = default_collate(batch)
#             # save collate result in dict
#             res = {}
#             for key, val in collate_res.items():
#                 val.labels = labels
#                 res[key] = val
#             __init__(self, location, tensor):
#             self._tensor = tensor
#             self._location = location
#             self._len = len(tensor)

#         def __getitem__(self, index):
#             tensor = self._tensor.select(0, index)
#             return {self._location: tensor}

#         def __len__(self):
#             return self._len


class LabelTensorDataLoader(DataLoader):

    def collate_fn(self, data):
        pass



"""Module for the TimeDependentProblem class"""
from abc import abstractmethod

from .abstract_problem import AbstractProblem


class TimeDependentProblem(AbstractProblem):
    """
    The class for the definition of time-dependent problems, i.e., for problems
    depending on time.

    Here's an example of a 1D wave problem.

    :Example:
        >>> from pina.problem import SpatialProblem, TimeDependentProblem
        >>> from pina.operators import grad, laplacian
        >>> from pina.equation import Equation, FixedValue
        >>> from pina import Condition
        >>> from pina.geometry import CartesianDomain
        >>> import torch
        >>>
        >>>
        >>> class Wave(TimeDependentSpatialProblem):
        >>>
        >>>     output_variables = ['u']
        >>>     spatial_domain = CartesianDomain({'x': [0, 3]})
        >>>     temporal_domain = CartesianDomain({'t': [0, 1]})
        >>>
        >>>     def wave_equation(input_, output_):
        >>>         u_t = grad(output_, input_, components=['u'], d=['t'])
        >>>         u_tt = grad(u_t, input_, components=['dudt'], d=['t'])
        >>>         delta_u = laplacian(output_, input_, components=['u'], d=['x'])
        >>>         return delta_u - u_tt
        >>>
        >>>     def initial_condition(input_, output_):
        >>>         u_expected = (-3*torch.sin(2*torch.pi*input_.extract(['x']))
        >>>             + 5*torch.sin(8/3*torch.pi*input_.extract(['x'])))
        >>>         u = output_.extract(['u'])
        >>>         return u - u_expected
        >>>
        >>>     conditions = {
        >>>         't0': Condition(CartesianDomain({'x': [0, 3], 't':0}), Equation(initial_condition)),
        >>>         'gamma1': Condition(CartesianDomain({'x':0, 't':[0, 1]}), FixedValue(0.)),
        >>>         'gamma2': Condition(CartesianDomain({'x':3, 't':[0, 1]}), FixedValue(0.)),
        >>>         'D': Condition(CartesianDomain({'x': [0, 3], 't':[0, 1]}), Equation(wave_equation))}
    """

    @abstractmethod
    def temporal_domain(self):
        """
        The temporal domain of the problem.
        """
        pass

    @property
    def temporal_variable(self):
        """
        The time variable of the problem.
        """
        return self.temporal_domain.variables



import torch


class Stride(object):

    def __init__(self, dict):
        """Stride class for continous convolution

        :param param: type of continuous convolution
        :type param: string
        """

        self._dict_stride = dict
        self._stride_continuous = None
        self._stride_discrete = self._create_stride_discrete(dict)

    def _create_stride_discrete(self, my_dict):
        """Creating the list for applying the filter

        :param my_dict: Dictionary with the following arguments:
            domain size, starting position of the filter, jump size
            for the filter and direction of the filter
        :type my_dict: dict
        :raises IndexError: Values in the dict must have all same length
        :raises ValueError: Domain values must be greater than 0
        :raises ValueError: Direction must be either equal to 1, -1 or 0
        :raises IndexError: Direction and jumps must have zero in the same
            index
        :return: list of positions for the filter
        :rtype: list
        :Example:


                >>> stride = {"domain": [4, 4],
                            "start": [-4, 2],
                            "jump": [2, 2],
                            "direction": [1, 1],
                            }
                >>> create_stride(stride)
                [[-4.0, 2.0], [-4.0, 4.0], [-2.0, 2.0], [-2.0, 4.0]]
        """

        # we must check boundaries of the input as well

        domain, start, jumps, direction = my_dict.values()

        # checking

        if not all([len(s) == len(domain) for s in my_dict.values()]):
            raise IndexError("values in the dict must have all same length")

        if not all(v >= 0 for v in domain):
            raise ValueError("domain values must be greater than 0")

        if not all(v == 1 or v == -1 or v == 0 for v in direction):
            raise ValueError("direction must be either equal to 1, -1 or 0")

        seq_jumps = [i for i, e in enumerate(jumps) if e == 0]
        seq_direction = [i for i, e in enumerate(direction) if e == 0]

        if seq_direction != seq_jumps:
            raise IndexError(
                "direction and jumps must have zero in the same index")

        if seq_jumps:
            for i in seq_jumps:
                jumps[i] = domain[i]
                direction[i] = 1

        # creating the stride grid
        values_mesh = [
            torch.arange(0, i, step).float() for i, step in zip(domain, jumps)
        ]

        values_mesh = [
            single * dim for single, dim in zip(values_mesh, direction)
        ]

        mesh = torch.meshgrid(values_mesh)
        coordinates_mesh = [x.reshape(-1, 1) for x in mesh]

        stride = torch.cat(coordinates_mesh, dim=1) + torch.tensor(start)

        return stride



"""Module for the ParametricProblem class"""
from abc import abstractmethod

from .abstract_problem import AbstractProblem


class InverseProblem(AbstractProblem):
    """
    The class for the definition of inverse problems, i.e., problems
    with unknown parameters that have to be learned during the training process
    from given data.

    Here's an example of a spatial inverse ODE problem, i.e., a spatial
    ODE problem with an unknown parameter `alpha` as coefficient of the
    derivative term.

    :Example:
        >>> from pina.problem import SpatialProblem, InverseProblem
        >>> from pina.operators import grad
        >>> from pina.equation import ParametricEquation, FixedValue
        >>> from pina import Condition
        >>> from pina.geometry import CartesianDomain
        >>> import torch
        >>>
        >>> class InverseODE(SpatialProblem, InverseProblem):
        >>>
        >>>     output_variables = ['u']
        >>>     spatial_domain = CartesianDomain({'x': [0, 1]})
        >>>     unknown_parameter_domain = CartesianDomain({'alpha': [1, 10]})
        >>>
        >>>     def ode_equation(input_, output_, params_):
        >>>         u_x = grad(output_, input_, components=['u'], d=['x'])
        >>>         u = output_.extract(['u'])
        >>>         return params_.extract(['alpha']) * u_x - u
        >>>
        >>>     def solution_data(input_, output_):
        >>>         x = input_.extract(['x'])
        >>>         solution = torch.exp(x)
        >>>         return output_ - solution
        >>>
        >>>     conditions = {
        >>>         'x0': Condition(CartesianDomain({'x': 0}), FixedValue(1.0)),
        >>>         'D': Condition(CartesianDomain({'x': [0, 1]}), ParametricEquation(ode_equation)),
        >>>         'data': Condition(CartesianDomain({'x': [0, 1]}), Equation(solution_data))
    """

    @abstractmethod
    def unknown_parameter_domain(self):
        """
        The parameters' domain of the problem.
        """
        pass

    @property
    def unknown_variables(self):
        """
        The parameters of the problem.
        """
        return self.unknown_parameter_domain.variables

    @property
    def unknown_parameters(self):
        """
        The parameters of the problem.
        """
        return self.__unknown_parameters

    @unknown_parameters.setter
    def unknown_parameters(self, value):
        self.__unknown_parameters = value




import torch
import torch.nn as nn
from ..utils import check_consistency
from .layers.fourier import FourierBlock1D, FourierBlock2D, FourierBlock3D
from pina import LabelTensor
import warnings


class FNO(torch.nn.Module):
    """
    The PINA implementation of Fourier Neural Operator network.

    Fourier Neural Operator (FNO) is a general architecture for learning Operators.
    Unlike traditional machine learning methods FNO is designed to map
    entire functions to other functions. It can be trained both with 
    Supervised learning strategies. FNO does global convolution by performing the
    operation on the Fourier space.

    .. seealso::

        **Original reference**: Li, Z., Kovachki, N., Azizzadenesheli, K., Liu, B.,
        Bhattacharya, K., Stuart, A., & Anandkumar, A. (2020). *Fourier neural operator for
        parametric partial differential equations*.
        DOI: `arXiv preprint arXiv:2010.08895.
        <https://arxiv.org/abs/2010.08895>`_
    """

    def __init__(self,
                 lifting_net,
                 projecting_net,
                 n_modes,
                 dimensions=3,
                 padding=8,
                 padding_type="constant",
                 inner_size=20,
                 n_layers=2,
                 func=nn.Tanh,
                 layers=None):
        super().__init__()

        # check type consistency
        check_consistency(lifting_net, nn.Module)
        check_consistency(projecting_net, nn.Module)
        check_consistency(dimensions, int)
        check_consistency(padding, int)
        check_consistency(padding_type, str)
        check_consistency(inner_size, int)
        check_consistency(n_layers, int)
        check_consistency(func, nn.Module, subclass=True)
        if layers is not None:
            if isinstance(layers, (tuple, list)):
                check_consistency(layers, int)
            else:
                raise ValueError('layers must be tuple or list of int.')
        if not isinstance(n_modes, (list, tuple, int)):
            raise ValueError(
                'n_modes must be a int or list or tuple of valid modes.'
                ' More information on the official documentation.')

        # assign variables
        # TODO check input lifting net and input projecting net
        self._lifting_net = lifting_net
        self._projecting_net = projecting_net
        self._padding = padding

        # initialize fourier layer for each dimension
        if dimensions == 1:
            fourier_layer = FourierBlock1D
        elif dimensions == 2:
            fourier_layer = FourierBlock2D
        elif dimensions == 3:
            fourier_layer = FourierBlock3D
        else:
            raise NotImplementedError('FNO implemented only for 1D/2D/3D data.')

        # Here we build the FNO by stacking Fourier Blocks

        # 1. Assign output dimensions for each FNO layer
        if layers is None:
            layers = [inner_size] * n_layers

        # 2. Assign activation functions for each FNO layer
        if isinstance(func, list):
            if len(layers) != len(func):
                raise RuntimeError(
                    'Uncosistent number of layers and functions.')
            self._functions = func
        else:
            self._functions = [func for _ in range(len(layers))]

        # 3. Assign modes functions for each FNO layer
        if isinstance(n_modes, list):
            if all(isinstance(i, list)
                   for i in n_modes) and len(layers) != len(n_modes):
                raise RuntimeError(
                    'Uncosistent number of layers and functions.')
            elif all(isinstance(i, int) for i in n_modes):
                n_modes = [n_modes] * len(layers)
        else:
            n_modes = [n_modes] * len(layers)

        # 4. Build the FNO network
        tmp_layers = layers.copy()
        first_parameter = next(lifting_net.parameters())
        input_shape = first_parameter.size()
        out_feats = lifting_net(torch.rand(size=input_shape)).shape[-1]
        tmp_layers.insert(0, out_feats)

        self._layers = []
        for i in range(len(tmp_layers) - 1):
            self._layers.append(
                fourier_layer(input_numb_fields=tmp_layers[i],
                              output_numb_fields=tmp_layers[i + 1],
                              n_modes=n_modes[i],
                              activation=self._functions[i]))
        self._layers = nn.Sequential(*self._layers)

        # 5. Padding values for spectral conv
        if isinstance(padding, int):
            padding = [padding] * dimensions
        self._ipad = [-pad if pad > 0 else None for pad in padding[:dimensions]]
        self._padding_type = padding_type
        self._pad = [
            val for pair in zip([0] * dimensions, padding) for val in pair
        ]

    def forward(self, x):
        """
        Forward computation for Fourier Neural Operator. It performs a
        lifting of the input by the ``lifting_net``. Then different layers
        of Fourier Blocks are applied. Finally the output is projected
        to the final dimensionality by the ``projecting_net``.

        :param torch.Tensor x: The input tensor for fourier block, depending on
            ``dimension`` in the initialization. In particular it is expected
            * 1D tensors: ``[batch, X, channels]``
            * 2D tensors: ``[batch, X, Y, channels]``
            * 3D tensors: ``[batch, X, Y, Z, channels]``
        :return: The output tensor obtained from the FNO.
        :rtype: torch.Tensor
        """
        if isinstance(x, LabelTensor): #TODO remove when Network is fixed
            warnings.warn('LabelTensor passed as input is not allowed, casting LabelTensor to Torch.Tensor')
            x = x.as_subclass(torch.Tensor)

        # lifting the input in higher dimensional space
        x = self._lifting_net(x)

        # permuting the input [batch, channels, x, y, ...]
        permutation_idx = [0, x.ndim - 1, *[i for i in range(1, x.ndim - 1)]]
        x = x.permute(permutation_idx)

        # padding the input
        x = torch.nn.functional.pad(x, pad=self._pad, mode=self._padding_type)

        # apply fourier layers
        x = self._layers(x)

        # remove padding
        idxs = [slice(None), slice(None)] + [slice(pad) for pad in self._ipad]
        x = x[idxs]

        # permuting back [batch, x, y, ..., channels]
        permutation_idx = [0, *[i for i in range(2, x.ndim)], 1]
        x = x.permute(permutation_idx)

        # apply projecting operator and return
        return self._projecting_net(x)



__all__ = [
    'AbstractProblem',
    'SpatialProblem',
    'TimeDependentProblem',
    'ParametricProblem',
    'InverseProblem',
]

from .abstract_problem import AbstractProblem
from .spatial_problem import SpatialProblem
from .timedep_problem import TimeDependentProblem
from .parametric_problem import ParametricProblem
from .inverse_problem import InverseProblem



'''PINA Callbacks Implementations'''

from pytorch_lightning.callbacks import Callback
import torch
import copy


class MetricTracker(Callback):
    
    def __init__(self):
        """
        PINA Implementation of a Lightning Callback for Metric Tracking.

        This class provides functionality to track relevant metrics during the training process.

        :ivar _collection: A list to store collected metrics after each training epoch.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer

        :return: A dictionary containing aggregated metric values.
        :rtype: dict

        Example:
            >>> tracker = MetricTracker()
            >>> # ... Perform training ...
            >>> metrics = tracker.metrics
        """
        self._collection = []

    def on_train_epoch_end(self, trainer, __):
        """
        Collect and track metrics at the end of each training epoch.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer
        :param _: Placeholder argument.

        :return: None
        :rtype: None
        """
        self._collection.append(copy.deepcopy(
            trainer.logged_metrics))  # track them

    @property
    def metrics(self):
        """
        Aggregate collected metrics during training.

        :return: A dictionary containing aggregated metric values.
        :rtype: dict
        """
        common_keys = set.intersection(*map(set, self._collection))
        v = {
            k: torch.stack([dic[k] for dic in self._collection])
            for k in common_keys
        }
        return v



from .adaptive_tanh import AdaptiveTanh
from .adaptive_sin import AdaptiveSin
from .adaptive_cos import AdaptiveCos
from .adaptive_linear import AdaptiveLinear
from .adaptive_square import AdaptiveSquare



""" Module for plotting. """
import matplotlib.pyplot as plt
import numpy as np
import torch

from pina import LabelTensor


class Writer:
    """
    Implementation of a writer class, for textual output.
    """

    def __init__(self, frequency_print=10, header='any') -> None:
        """
        The constructor of the class.

        :param int frequency_print: the frequency in epochs of printing.
        :param ['any', 'begin', 'none'] header: the header of the output.
        """

        self._frequency_print = frequency_print
        self._header = header

    def header(self, trainer):
        """
        The method for printing the header.
        """
        header = []
        for condition_name in trainer.problem.conditions:
            header.append(f'{condition_name}')

        return header

    def write_loss(self, trainer):
        """
        The method for writing the output.
        """
        pass

    def write_loss_in_loop(self, trainer, loss):
        """
        The method for writing the output within the training loop.

        :param pina.trainer.Trainer trainer: the trainer object.
        """

        if trainer.trained_epoch % self._frequency_print == 0:
            print(f'Epoch {trainer.trained_epoch:05d}: {loss.item():.5e}')



import torch

from .location import Location
from ..label_tensor import LabelTensor
from ..utils import check_consistency


class EllipsoidDomain(Location):
    """PINA implementation of Ellipsoid domain."""

    def __init__(self, ellipsoid_dict, sample_surface=False):
        """PINA implementation of Ellipsoid domain.

        :param ellipsoid_dict: A dictionary with dict-key a string representing
            the input variables for the pinn, and dict-value a list with
            the domain extrema.
        :type ellipsoid_dict: dict
        :param sample_surface: A variable for choosing sample strategies. If
            ``sample_surface=True`` only samples on the ellipsoid surface
            frontier are taken. If ``sample_surface=False`` only samples on
            the ellipsoid interior are taken, defaults to ``False``.
        :type sample_surface: bool

        .. warning::
            Sampling for dimensions greater or equal to 10 could result
            in a shrinking of the ellipsoid, which degrades the quality
            of the samples. For dimensions higher than 10, other algorithms
            for sampling should be used, such as: Dezert, Jean, and Christian
            Musso. "An efficient method for generating points uniformly
            distributed in hyperellipsoids." Proceedings of the Workshop on
            Estimation, Tracking and Fusion: A Tribute to Yaakov Bar-Shalom.
            Vol. 7. No. 8. 2001.

        :Example:
            >>> spatial_domain = Ellipsoid({'x':[-1, 1], 'y':[-1,1]})

        """
        self.fixed_ = {}
        self.range_ = {}
        self._centers = None
        self._axis = None

        # checking consistency
        check_consistency(sample_surface, bool)
        self._sample_surface = sample_surface

        for k, v in ellipsoid_dict.items():
            if isinstance(v, (int, float)):
                self.fixed_[k] = v
            elif isinstance(v, (list, tuple)) and len(v) == 2:
                self.range_[k] = v
            else:
                raise TypeError

        # perform operation only for not fixed variables (if any)

        if self.range_:

            # convert dict vals to torch [dim, 2] matrix
            list_dict_vals = list(self.range_.values())
            tmp = torch.tensor(list_dict_vals, dtype=torch.float)

            # get the ellipsoid center
            normal_basis = torch.eye(len(list_dict_vals))
            centers = torch.diag(normal_basis * tmp.mean(axis=1))

            # get the ellipsoid axis
            ellipsoid_axis = (tmp - centers.reshape(-1, 1))[:, -1]

            # save elipsoid axis and centers as dict
            self._centers = dict(zip(self.range_.keys(), centers.tolist()))
            self._axis = dict(zip(self.range_.keys(), ellipsoid_axis.tolist()))

    @property
    def variables(self):
        """Spatial variables.

        :return: Spatial variables defined in '__init__()'
        :rtype: list[str]
        """
        return list(self.fixed_.keys()) + list(self.range_.keys())

    def is_inside(self, point, check_border=False):
        """Check if a point is inside the ellipsoid domain.

        .. note::
            When ``sample_surface`` in the ``__init()__``
            is set to ``True``, then the method only checks 
            points on the surface, and not inside the domain.

        :param point: Point to be checked.
        :type point: LabelTensor
        :param check_border: Check if the point is also on the frontier
            of the ellipsoid, default ``False``.
        :type check_border: bool
        :return: Returning True if the point is inside, ``False`` otherwise.
        :rtype: bool
        """

        # small check that point is labeltensor
        check_consistency(point, LabelTensor)

        # get axis ellipse as tensors
        list_dict_vals = list(self._axis.values())
        tmp = torch.tensor(list_dict_vals, dtype=torch.float)
        ax_sq = LabelTensor(tmp.reshape(1, -1)**2, self.variables)

        # get centers ellipse as tensors
        list_dict_vals = list(self._centers.values())
        tmp = torch.tensor(list_dict_vals, dtype=torch.float)
        centers = LabelTensor(tmp.reshape(1, -1), self.variables)

        if not all([i in ax_sq.labels for i in point.labels]):
            raise ValueError('point labels different from constructor'
                             f' dictionary labels. Got {point.labels},'
                             f' expected {ax_sq.labels}.')

        # point square + shift center
        point_sq = (point - centers).pow(2)
        point_sq.labels = point.labels

        # calculate ellispoid equation
        eqn = torch.sum(point_sq.extract(ax_sq.labels) / ax_sq) - 1.

        # if we have sampled only the surface, we check that the
        # point is inside the surface border only
        if self._sample_surface:
            return torch.allclose(eqn, torch.zeros_like(eqn))

        # otherwise we check the ellipse
        if check_border:
            return bool(eqn <= 0)

        return bool(eqn < 0)

    def _sample_range(self, n, mode, variables):
        """Rescale the samples to the correct bounds.

        :param n: Number of points to sample in the ellipsoid.
        :type n: int
        :param mode: Mode for sampling, defaults to ``random``.
            Available modes include: random sampling, ``random``.
        :type mode: str, optional
        :param variables: Variables to  be rescaled in the samples.
        :type variables: torch.Tensor
        :return: Rescaled sample points.
        :rtype: torch.Tensor
        """

        # =============== For Developers ================ #
        #
        # The sampling startegy used is fairly simple.
        # For all `mode`s first we sample from the unit
        # sphere and then we scale and shift according
        # to self._axis.values() and self._centers.values().
        #
        # =============================================== #

        # get dimension
        dim = len(variables)

        # get values center
        pairs_center = [(k, v) for k, v in self._centers.items()
                        if k in variables]
        _, values_center = map(list, zip(*pairs_center))
        values_center = torch.tensor(values_center)

        # get values axis
        pairs_axis = [(k, v) for k, v in self._axis.items() if k in variables]
        _, values_axis = map(list, zip(*pairs_axis))
        values_axis = torch.tensor(values_axis)

        # Sample in the unit sphere
        if mode == 'random':
            # 1. Sample n points from the surface of a unit sphere
            # 2. Scale each dimension using torch.rand()
            #    (a random number between 0-1) so that it lies within
            #    the sphere, only if self._sample_surface=False
            # 3. Multiply with self._axis.values() to make it ellipsoid
            # 4. Shift the mean of the ellipse by adding self._centers.values()

            # step 1.
            pts = torch.randn(size=(n, dim))
            pts = pts / torch.linalg.norm(pts, axis=-1).view((n, 1))
            if not self._sample_surface:  # step 2.
                scale = torch.rand((n, 1))
                pts = pts * scale

            # step 3. and 4.
            pts *= values_axis
            pts += values_center

        return pts

    def sample(self, n, mode='random', variables='all'):
        """Sample routine.

        :param int n: Number of points to sample in the shape.
        :param str mode: Mode for sampling, defaults to ``random``. Available modes include: ``random``.
        :param variables: Variables to be sampled, defaults to ``all``.
        :type variables: str | list[str]
        :return: Returns ``LabelTensor`` of n sampled points.
        :rtype: LabelTensor

        :Example:
            >>> elips = Ellipsoid({'x':[1, 0], 'y':1})
            >>> elips.sample(n=6)
                tensor([[0.4872, 1.0000],
                        [0.2977, 1.0000],
                        [0.0422, 1.0000],
                        [0.6431, 1.0000],
                        [0.7272, 1.0000],
                        [0.8326, 1.0000]])
        """

        def _Nd_sampler(n, mode, variables):
            """Sample all the variables together

            :param n: Number of points to sample.
            :type n: int
            :param mode: Mode for sampling, defaults to ``random``.
                Available modes include: random sampling, ``random``;
                latin hypercube sampling, 'latin' or 'lh';
                chebyshev sampling, 'chebyshev'; grid sampling 'grid'.
            :type mode: str, optional.
            :param variables: pinn variable to be sampled, defaults to ``all``.
            :type variables: str or list[str], optional.
            :return: Sample points.
            :rtype: list[torch.Tensor]
            """
            pairs = [(k, v) for k, v in self.range_.items() if k in variables]
            keys, _ = map(list, zip(*pairs))

            result = self._sample_range(n, mode, keys)
            result = result.as_subclass(LabelTensor)
            result.labels = keys

            for variable in variables:
                if variable in self.fixed_.keys():
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]
                                                 ]).repeat(result.shape[0], 1)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]

                    result = result.append(pts_variable, mode='std')
            return result

        def _single_points_sample(n, variables):
            """Sample a single point in one dimension.

            :param n: Number of points to sample.
            :type n: int
            :param variables: Variables to sample from.
            :type variables: list[str]
            :return: Sample points.
            :rtype: list[torch.Tensor]
            """
            tmp = []
            for variable in variables:
                if variable in self.fixed_.keys():
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]]).repeat(n, 1)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]
                    tmp.append(pts_variable)

            result = tmp[0]
            for i in tmp[1:]:
                result = result.append(i, mode='std')

            return result

        if self.fixed_ and (not self.range_):
            return _single_points_sample(n, variables)

        if variables == 'all':
            variables = list(self.range_.keys()) + list(self.fixed_.keys())

        if mode in ['random']:
            return _Nd_sampler(n, mode, variables)
        else:
            raise NotImplementedError(f'mode={mode} is not implemented.')



""" Module for SupervisedSolver """
import torch
try:
    from torch.optim.lr_scheduler import LRScheduler  # torch >= 2.0
except ImportError:
    from torch.optim.lr_scheduler import _LRScheduler as LRScheduler  # torch < 2.0

from torch.optim.lr_scheduler import ConstantLR

from .solver import SolverInterface
from ..label_tensor import LabelTensor
from ..utils import check_consistency
from ..loss import LossInterface
from torch.nn.modules.loss import _Loss


class SupervisedSolver(SolverInterface):
    """
    SupervisedSolver solver class. This class implements a SupervisedSolver,
    using a user specified ``model`` to solve a specific ``problem``. 
    """

    def __init__(
        self,
        problem,
        model,
        extra_features=None,
        loss=torch.nn.MSELoss(),
        optimizer=torch.optim.Adam,
        optimizer_kwargs={'lr': 0.001},
        scheduler=ConstantLR,
        scheduler_kwargs={
            "factor": 1,
            "total_iters": 0
        },
    ):
        '''
        :param AbstractProblem problem: The formualation of the problem.
        :param torch.nn.Module model: The neural network model to use.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.nn.Module extra_features: The additional input
            features to use as augmented input.
        :param torch.optim.Optimizer optimizer: The neural network optimizer to
            use; default is :class:`torch.optim.Adam`.
        :param dict optimizer_kwargs: Optimizer constructor keyword args.
        :param float lr: The learning rate; default is 0.001.
        :param torch.optim.LRScheduler scheduler: Learning
            rate scheduler.
        :param dict scheduler_kwargs: LR scheduler constructor keyword args.
        '''
        super().__init__(models=[model],
                         problem=problem,
                         optimizers=[optimizer],
                         optimizers_kwargs=[optimizer_kwargs],
                         extra_features=extra_features)

        # check consistency
        check_consistency(scheduler, LRScheduler, subclass=True)
        check_consistency(scheduler_kwargs, dict)
        check_consistency(loss, (LossInterface, _Loss), subclass=False)

        # assign variables
        self._scheduler = scheduler(self.optimizers[0], **scheduler_kwargs)
        self._loss = loss
        self._neural_net = self.models[0]

    def forward(self, x):
        """Forward pass implementation for the solver.

        :param torch.Tensor x: Input tensor. 
        :return: Solver solution.
        :rtype: torch.Tensor
        """
        return self.neural_net(x)

    def configure_optimizers(self):
        """Optimizer configuration for the solver.

        :return: The optimizers and the schedulers
        :rtype: tuple(list, list)
        """
        return self.optimizers, [self.scheduler]

    def training_step(self, batch, batch_idx):
        """Solver training step.

        :param batch: The batch element in the dataloader.
        :type batch: tuple
        :param batch_idx: The batch index.
        :type batch_idx: int
        :return: The sum of the loss functions.
        :rtype: LabelTensor
        """

        dataloader = self.trainer.train_dataloader
        condition_idx = batch['condition']

        for condition_id in range(condition_idx.min(), condition_idx.max()+1):

            condition_name = dataloader.condition_names[condition_id]
            condition = self.problem.conditions[condition_name]
            pts = batch['pts']
            out = batch['output']

            if condition_name not in self.problem.conditions:
                raise RuntimeError('Something wrong happened.')

            # for data driven mode
            if not hasattr(condition, 'output_points'):
                raise NotImplementedError('Supervised solver works only in data-driven mode.')
            
            output_pts = out[condition_idx == condition_id]
            input_pts = pts[condition_idx == condition_id]

            loss = self.loss(self.forward(input_pts), output_pts) * condition.data_weight
            loss = loss.as_subclass(torch.Tensor)

        self.log('mean_loss', float(loss), prog_bar=True, logger=True)
        return loss

    @property
    def scheduler(self):
        """
        Scheduler for training.
        """
        return self._scheduler

    @property
    def neural_net(self):
        """
        Neural network for training.
        """
        return self._neural_net

    @property
    def loss(self):
        """
        Loss for training.
        """
        return self._loss



""" Module for AbstractProblem class """
from abc import ABCMeta, abstractmethod
from ..utils import merge_tensors, check_consistency
import torch


class AbstractProblem(metaclass=ABCMeta):
    """
    The abstract `AbstractProblem` class. All the class defining a PINA Problem
    should be inheritied from this class.

    In the definition of a PINA problem, the fundamental elements are:
    the output variables, the condition(s), and the domain(s) where the
    conditions are applied.
    """

    def __init__(self):

        # variable storing all points
        self.input_pts = {}

        # varible to check if sampling is done. If no location
        # element is presented in Condition this variable is set to true
        self._have_sampled_points = {}
        for condition_name in self.conditions:
            self._have_sampled_points[condition_name] = False

        # put in self.input_pts all the points that we don't need to sample
        self._span_condition_points()

    @property
    def input_variables(self):
        """
        The input variables of the AbstractProblem, whose type depends on the
        type of domain (spatial, temporal, and parameter).

        :return: the input variables of self
        :rtype: list

        """
        variables = []

        if hasattr(self, 'spatial_variables'):
            variables += self.spatial_variables
        if hasattr(self, 'temporal_variable'):
            variables += self.temporal_variable
        if hasattr(self, 'parameters'):
            variables += self.parameters
        if hasattr(self, 'custom_variables'):
            variables += self.custom_variables

        return variables

    @property
    def domain(self):
        """
        The domain(s) where the conditions of the AbstractProblem are valid.
        If more than one domain type is passed, a list of Location is
        retured.

        :return: the domain(s) of ``self``
        :rtype: list[Location]
        """
        domains = [
            getattr(self, f'{t}_domain')
            for t in ['spatial', 'temporal', 'parameter']
            if hasattr(self, f'{t}_domain')
        ]

        if len(domains) == 1:
            return domains[0]
        elif len(domains) == 0:
            raise RuntimeError

        if len(set(map(type, domains))) == 1:
            domain = domains[0].__class__({})
            [domain.update(d) for d in domains]
            return domain
        else:
            raise RuntimeError('different domains')

    @input_variables.setter
    def input_variables(self, variables):
        raise RuntimeError

    @property
    @abstractmethod
    def output_variables(self):
        """
        The output variables of the problem.
        """
        pass

    @property
    @abstractmethod
    def conditions(self):
        """
        The conditions of the problem.
        """
        pass

    def _span_condition_points(self):
        """
        Simple function to get the condition points
        """
        for condition_name in self.conditions:
            condition = self.conditions[condition_name]
            if hasattr(condition, 'input_points'):
                samples = condition.input_points
                self.input_pts[condition_name] = samples
                self._have_sampled_points[condition_name] = True
            if hasattr(self, 'unknown_parameter_domain'):
                # initialize the unknown parameters of the inverse problem given
                # the domain the user gives
                self.unknown_parameters = {}
                for i, var in enumerate(self.unknown_variables):
                    range_var = self.unknown_parameter_domain.range_[var]
                    tensor_var = torch.rand(1, requires_grad=True) * range_var[1] + range_var[0]
                    self.unknown_parameters[var] = torch.nn.Parameter(tensor_var)

    def discretise_domain(self,
                          n,
                          mode='random',
                          variables='all',
                          locations='all'):
        """
        Generate a set of points to span the `Location` of all the conditions of
        the problem.

        :param n: Number of points to sample, see Note below
            for reference.
        :type n: int
        :param mode: Mode for sampling, defaults to ``random``.
            Available modes include: random sampling, ``random``;
            latin hypercube sampling, ``latin`` or ``lh``;
            chebyshev sampling, ``chebyshev``; grid sampling ``grid``.
        :param variables: problem's variables to be sampled, defaults to 'all'.
        :type variables: str | list[str]
        :param locations: problem's locations from where to sample, defaults to 'all'.
        :type locations: str

        :Example:
            >>> pinn.discretise_domain(n=10, mode='grid')
            >>> pinn.discretise_domain(n=10, mode='grid', location=['bound1'])
            >>> pinn.discretise_domain(n=10, mode='grid', variables=['x'])

        .. warning::
            ``random`` is currently the only implemented ``mode`` for all geometries, i.e.
            ``EllipsoidDomain``, ``CartesianDomain``, ``SimplexDomain`` and the geometries
            compositions ``Union``, ``Difference``, ``Exclusion``, ``Intersection``. The
            modes ``latin`` or ``lh``,  ``chebyshev``, ``grid`` are only implemented for
            ``CartesianDomain``.
        """

        # check consistecy n
        check_consistency(n, int)

        # check consistency mode
        check_consistency(mode, str)
        if mode not in ['random', 'grid', 'lh', 'chebyshev', 'latin']:
            raise TypeError(f'mode {mode} not valid.')

        # check consistency variables
        if variables == 'all':
            variables = self.input_variables
        else:
            check_consistency(variables, str)

        if sorted(variables) != sorted(self.input_variables):
            TypeError(f'Wrong variables for sampling. Variables ',
                      f'should be in {self.input_variables}.')

        # check consistency location
        if locations == 'all':
            locations = [condition for condition in self.conditions]
        else:
            check_consistency(locations, str)

        if sorted(locations) != sorted(self.conditions):
            TypeError(f'Wrong locations for sampling. Location ',
                      f'should be in {self.conditions}.')

        # sampling
        for location in locations:
            condition = self.conditions[location]

            # we try to check if we have already sampled
            try:
                already_sampled = [self.input_pts[location]]
            # if we have not sampled, a key error is thrown
            except KeyError:
                already_sampled = []

            # if we have already sampled fully the condition
            # but we want to sample again we set already_sampled
            # to an empty list since we need to sample again, and
            # self._have_sampled_points to False.
            if self._have_sampled_points[location]:
                already_sampled = []
                self._have_sampled_points[location] = False

            # build samples
            samples = [
                condition.location.sample(n=n, mode=mode, variables=variables)
            ] + already_sampled
            pts = merge_tensors(samples)
            self.input_pts[location] = pts

            # the condition is sampled if input_pts contains all labels
            if sorted(self.input_pts[location].labels) == sorted(
                    self.input_variables):
                self._have_sampled_points[location] = True


    def add_points(self, new_points):
        """
        Adding points to the already sampled points.

        :param dict new_points: a dictionary with key the location to add the points
            and values the torch.Tensor points.
        """

        if sorted(new_points.keys()) != sorted(self.conditions):
            TypeError(f'Wrong locations for new points. Location ',
                      f'should be in {self.conditions}.')

        for location in new_points.keys():
            # extract old and new points
            old_pts = self.input_pts[location]
            new_pts = new_points[location]

            # if they don't have the same variables error
            if sorted(old_pts.labels) != sorted(new_pts.labels):
                TypeError(f'Not matching variables for old and new points '
                          f'in condition {location}.')
            if old_pts.labels != new_pts.labels:
                new_pts = torch.hstack(
                    [new_pts.extract([i]) for i in old_pts.labels])
                new_pts.labels = old_pts.labels

            # merging
            merged_pts = torch.vstack([old_pts, new_points[location]])
            merged_pts.labels = old_pts.labels
            self.input_pts[location] = merged_pts

    @property
    def have_sampled_points(self):
        """
        Check if all points for
        ``Location`` are sampled.
        """
        return all(self._have_sampled_points.values())

    @property
    def not_sampled_points(self):
        """
        Check which points are
        not sampled.
        """
        # variables which are not sampled
        not_sampled = None
        if self.have_sampled_points is False:
            # check which one are not sampled:
            not_sampled = []
            for condition_name, is_sample in self._have_sampled_points.items():
                if not is_sample:
                    not_sampled.append(condition_name)
        return not_sampled




import torch
import torch.nn as nn
from ..utils import check_consistency
from ..label_tensor import LabelTensor


class Network(torch.nn.Module):

    def __init__(self, model, input_variables, output_variables, extra_features=None):
        """
        Network class with standard forward method
        and possibility to pass extra features. This
        class is used internally in PINA to convert
        any :class:`torch.nn.Module` s in a PINA module.
    
        :param model: The torch model to convert in a PINA model.
        :type model: torch.nn.Module
        :param list(str) input_variables: The input variables of the :class:`AbstractProblem`, whose type depends on the
            type of domain (spatial, temporal, and parameter).
        :param list(str) output_variables: The output variables of the :class:`AbstractProblem`, whose type depends on the
            problem setting.
        :param extra_features: List of torch models to augment the input, defaults to None.
        :type extra_features: list(torch.nn.Module)
        """
        super().__init__()

        # check model consistency
        check_consistency(model, nn.Module)
        check_consistency(input_variables, str)
        check_consistency(output_variables, str)

        self._model = model
        self._input_variables = input_variables
        self._output_variables = output_variables

        # check consistency and assign extra fatures
        if extra_features is None:
            self._extra_features = []
        else:
            for feat in extra_features:
                check_consistency(feat, nn.Module)
            self._extra_features = nn.Sequential(*extra_features)

        # check model works with inputs
        # TODO

    def forward(self, x):
        """
        Forward method for Network class. This class
        implements the standard forward method, and
        it adds the possibility to pass extra features.
        All the PINA models ``forward`` s are overriden
        by this class, to enable :class:`pina.label_tensor.LabelTensor` labels
        extraction.

        :param torch.Tensor x: Input of the network.
        :return torch.Tensor: Output of the network.
        """
        # only labeltensors as input
        assert isinstance(x, LabelTensor), "Expected LabelTensor as input to the model."

        # extract torch.Tensor from corresponding label
        # in case `input_variables = []` all points are used
        if self._input_variables:
            x = x.extract(self._input_variables)

        # extract features and append
        for feature in self._extra_features:
            x = x.append(feature(x))

        # perform forward pass + converting to LabelTensor
        output = self._model(x).as_subclass(LabelTensor)

        # set the labels for LabelTensor
        output.labels = self._output_variables

        return output
    
    # TODO to remove in next releases (only used in GAROM solver)
    def forward_map(self, x):
        """
        Forward method for Network class when the input is
        a tuple. This class is simply a forward with the input casted as a
        tuple or list :class`torch.Tensor`.
        All the PINA models ``forward`` s are overriden
        by this class, to enable :class:`pina.label_tensor.LabelTensor` labels
        extraction.

        :param list (torch.Tensor) | tuple(torch.Tensor) x: Input of the network.
        :return torch.Tensor: Output of the network.

        .. note::
            This function does not extract the input variables, all the variables
            are used for both tensors. Output variables are correctly applied.
        """
        # convert LabelTensor s to torch.Tensor s
        x = list(map(lambda x: x.as_subclass(torch.Tensor), x))

        # perform forward pass (using torch.Tensor) + converting to LabelTensor
        output = self._model(x).as_subclass(LabelTensor)

        # set the labels for LabelTensor
        output.labels = self._output_variables

        return output

    @property
    def torchmodel(self):
        return self._model

    @property
    def extra_features(self):
        return self._extra_features



import torch


class Integral(object):

    def __init__(self, param):
        """Integral class for continous convolution

        :param param: type of continuous convolution
        :type param: string
        """

        if param == 'discrete':
            self.make_integral = self.integral_param_disc
        elif param == 'continuous':
            self.make_integral = self.integral_param_cont
        else:
            raise TypeError

    def __call__(self, *args, **kwds):
        return self.make_integral(*args, **kwds)

    def _prepend_zero(self, x):
        """Create bins for performing integral

        :param x: input tensor
        :type x: torch.tensor
        :return: bins for integrals
        :rtype: torch.tensor
        """
        return torch.cat((torch.zeros(1, dtype=x.dtype, device=x.device), x))

    def integral_param_disc(self, x, y, idx):
        """Perform discretize integral
            with discrete parameters

        :param x: input vector
        :type x: torch.tensor
        :param y: input vector
        :type y: torch.tensor
        :param idx: indeces for different strides
        :type idx: list
        :return: integral
        :rtype: torch.tensor
        """
        cs_idxes = self._prepend_zero(torch.cumsum(torch.tensor(idx), 0))
        cs = self._prepend_zero(torch.cumsum(x.flatten() * y.flatten(), 0))
        return cs[cs_idxes[1:]] - cs[cs_idxes[:-1]]

    def integral_param_cont(self, x, y, idx):
        """Perform discretize integral for continuous convolution
            with continuous parameters

        :param x: input vector
        :type x: torch.tensor
        :param y: input vector
        :type y: torch.tensor
        :param idx: indeces for different strides
        :type idx: list
        :return: integral
        :rtype: torch.tensor
        """
        raise NotImplementedError



""" Module for Equation. """
from .equation_interface import EquationInterface


class Equation(EquationInterface):

    def __init__(self, equation):
        """
        Equation class for specifing any equation in PINA.
        Each ``equation`` passed to a ``Condition`` object
        must be an ``Equation`` or ``SystemEquation``.

        :param equation: A ``torch`` callable equation to
            evaluate the residual.
        :type equation: Callable
        """
        if not callable(equation):
            raise ValueError('equation must be a callable function.'
                             'Expected a callable function, got '
                             f'{equation}')
        self.__equation = equation

    def residual(self, input_, output_, params_ = None):
        """
        Residual computation of the equation.

        :param LabelTensor input_: Input points to evaluate the equation.
        :param LabelTensor output_: Output vectors given by a model (e.g,
            a ``FeedForward`` model).
        :param dict params_: Dictionary of parameters related to the inverse
            problem (if any).
            If the equation is not related to an ``InverseProblem``, the
            parameters are initialized to ``None`` and the residual is
            computed as ``equation(input_, output_)``.
            Otherwise, the parameters are automatically initialized in the
            ranges specified by the user.

        :return: The residual evaluation of the specified equation.
        :rtype: LabelTensor
        """
        if params_ is None:
            result = self.__equation(input_, output_)
        else:
            result = self.__equation(input_, output_, params_)
        return result



import torch

from pina.utils import merge_tensors
from pina.label_tensor import LabelTensor
from pina import LabelTensor
from pina.geometry import EllipsoidDomain, CartesianDomain
from pina.utils import check_consistency
import pytest
from pina.geometry import Location


def test_merge_tensors():
    tensor1 = LabelTensor(torch.rand((20, 3)), ['a', 'b', 'c'])
    tensor2 = LabelTensor(torch.zeros((20, 3)), ['d', 'e', 'f'])
    tensor3 = LabelTensor(torch.ones((30, 3)), ['g', 'h', 'i'])

    merged_tensor = merge_tensors((tensor1, tensor2, tensor3))
    assert tuple(merged_tensor.labels) == ('a', 'b', 'c', 'd', 'e', 'f', 'g',
                                           'h', 'i')
    assert merged_tensor.shape == (20 * 20 * 30, 9)
    assert torch.all(merged_tensor.extract(('d', 'e', 'f')) == 0)
    assert torch.all(merged_tensor.extract(('g', 'h', 'i')) == 1)


def test_check_consistency_correct():
    ellipsoid1 = EllipsoidDomain({'x': [1, 2], 'y': [-2, 1]})
    example_input_pts = LabelTensor(torch.tensor([[0, 0, 0]]), ['x', 'y', 'z'])

    check_consistency(example_input_pts, torch.Tensor)
    check_consistency(CartesianDomain, Location, subclass=True)
    check_consistency(ellipsoid1, Location)


def test_check_consistency_incorrect():
    ellipsoid1 = EllipsoidDomain({'x': [1, 2], 'y': [-2, 1]})
    example_input_pts = LabelTensor(torch.tensor([[0, 0, 0]]), ['x', 'y', 'z'])

    with pytest.raises(ValueError):
        check_consistency(example_input_pts, Location)
    with pytest.raises(ValueError):
        check_consistency(torch.Tensor, Location, subclass=True)
    with pytest.raises(ValueError):
        check_consistency(ellipsoid1, torch.Tensor)



__all__ = [
    'Location', 'CartesianDomain', 'EllipsoidDomain', 'Union', 'Intersection',
    'Exclusion', 'Difference', 'OperationInterface', 'SimplexDomain'
]

from .location import Location
from .cartesian import CartesianDomain
from .ellipsoid import EllipsoidDomain
from .exclusion_domain import Exclusion
from .intersection_domain import Intersection
from .union_domain import Union
from .difference_domain import Difference
from .operation_interface import OperationInterface
from .simplex import SimplexDomain



from torch.utils.data import Dataset
import torch
from pina import LabelTensor


class SamplePointDataset(Dataset):
    """
    This class is used to create a dataset of sample points.
    """

    def __init__(self, problem, device) -> None:
        """
        :param dict input_pts: The input points.
        """
        super().__init__()
        pts_list = []
        self.condition_names = []

        for name, condition in problem.conditions.items():
            if not hasattr(condition, 'output_points'):
                pts_list.append(problem.input_pts[name])
                self.condition_names.append(name)

        self.pts = LabelTensor.vstack(pts_list)

        if self.pts != []:
            self.condition_indeces = torch.cat([
                torch.tensor([i]*len(pts_list[i]))
                for i in range(len(self.condition_names))
            ], dim=0)
        else: # if there are no sample points
            self.condition_indeces = torch.tensor([])
            self.pts = torch.tensor([])

        self.pts = self.pts.to(device)
        self.condition_indeces = self.condition_indeces.to(device)
        
    def __len__(self):
        return self.pts.shape[0]
    

class DataPointDataset(Dataset):

    def __init__(self, problem, device) -> None:
        super().__init__()
        input_list = []
        output_list = [] 
        self.condition_names = []

        for name, condition in problem.conditions.items():
            if hasattr(condition, 'output_points'):
                input_list.append(problem.conditions[name].input_points)
                output_list.append(problem.conditions[name].output_points)
                self.condition_names.append(name)

        self.input_pts = LabelTensor.vstack(input_list)
        self.output_pts = LabelTensor.vstack(output_list)

        if self.input_pts != []:
            self.condition_indeces = torch.cat([
                torch.tensor([i]*len(input_list[i]))
                for i in range(len(self.condition_names))
            ], dim=0)
        else: # if there are no data points
            self.condition_indeces = torch.tensor([])
            self.input_pts = torch.tensor([])
            self.output_pts = torch.tensor([])

        self.input_pts = self.input_pts.to(device)
        self.output_pts = self.output_pts.to(device)
        self.condition_indeces = self.condition_indeces.to(device)

    def __len__(self):
        return self.input_pts.shape[0]


class SamplePointLoader:
    """
    This class is used to create a dataloader to use during the training.

    :var condition_names: The names of the conditions. The order is consistent
        with the condition indeces in the batches.
    :vartype condition_names: list[str]
    """

    def __init__(self, sample_dataset, data_dataset, batch_size=None, shuffle=True) -> None:
        """
        Constructor.

        :param SamplePointDataset sample_pts: The sample points dataset.
        :param int batch_size: The batch size. If ``None``, the batch size is
            set to the number of sample points. Default is ``None``.
        :param bool shuffle: If ``True``, the sample points are shuffled.
            Default is ``True``.
        """
        if not isinstance(sample_dataset, SamplePointDataset):
            raise TypeError(f'Expected SamplePointDataset, got {type(sample_dataset)}')
        if not isinstance(data_dataset, DataPointDataset):
            raise TypeError(f'Expected DataPointDataset, got {type(data_dataset)}')

        self.n_data_conditions = len(data_dataset.condition_names)
        self.n_phys_conditions = len(sample_dataset.condition_names)
        data_dataset.condition_indeces += self.n_phys_conditions

        self._prepare_sample_dataset(sample_dataset, batch_size, shuffle)
        self._prepare_data_dataset(data_dataset, batch_size, shuffle)

        self.condition_names = (
            sample_dataset.condition_names + data_dataset.condition_names)

        self.batch_list = []
        for i in range(len(self.batch_sample_pts)):
            self.batch_list.append(
                ('sample', i)
            )

        for i in range(len(self.batch_input_pts)):
            self.batch_list.append(
                ('data', i)
            )

        if shuffle:
            self.random_idx = torch.randperm(len(self.batch_list))   
        else:
            self.random_idx = torch.arange(len(self.batch_list))


    def _prepare_data_dataset(self, dataset, batch_size, shuffle):
        """
        Prepare the dataset for data points.

        :param SamplePointDataset dataset: The dataset.
        :param int batch_size: The batch size.
        :param bool shuffle: If ``True``, the sample points are shuffled.
        """
        self.sample_dataset = dataset

        if len(dataset) == 0:
            self.batch_data_conditions = []
            self.batch_input_pts = []
            self.batch_output_pts = []
            return

        if batch_size is None:
            batch_size = len(dataset)
        batch_num = len(dataset) // batch_size
        if len(dataset) % batch_size != 0:
            batch_num += 1

        output_labels = dataset.output_pts.labels
        input_labels = dataset.input_pts.labels
        self.tensor_conditions = dataset.condition_indeces

        if shuffle:
            idx = torch.randperm(dataset.input_pts.shape[0])
            self.input_pts = dataset.input_pts[idx]
            self.output_pts = dataset.output_pts[idx]
            self.tensor_conditions = dataset.condition_indeces[idx]

        self.batch_input_pts = torch.tensor_split(
            dataset.input_pts, batch_num)
        self.batch_output_pts = torch.tensor_split(
            dataset.output_pts, batch_num)

        for i in range(len(self.batch_input_pts)):
            self.batch_input_pts[i].labels = input_labels
            self.batch_output_pts[i].labels = output_labels
        
        self.batch_data_conditions = torch.tensor_split(
            self.tensor_conditions, batch_num)

    def _prepare_sample_dataset(self, dataset, batch_size, shuffle):
        """
        Prepare the dataset for sample points.

        :param DataPointDataset dataset: The dataset.
        :param int batch_size: The batch size.
        :param bool shuffle: If ``True``, the sample points are shuffled.
        """

        self.sample_dataset = dataset
        if len(dataset) == 0:
            self.batch_sample_conditions = []
            self.batch_sample_pts = []
            return

        if batch_size is None:
            batch_size = len(dataset)

        batch_num = len(dataset) // batch_size
        if len(dataset) % batch_size != 0:
            batch_num += 1
        
        self.tensor_pts = dataset.pts
        self.tensor_conditions = dataset.condition_indeces

        # if shuffle:
        #     idx = torch.randperm(self.tensor_pts.shape[0])
        #     self.tensor_pts = self.tensor_pts[idx]
        #     self.tensor_conditions = self.tensor_conditions[idx]
        
        self.batch_sample_pts = torch.tensor_split(self.tensor_pts, batch_num)
        for i in range(len(self.batch_sample_pts)):
            self.batch_sample_pts[i].labels = dataset.pts.labels

        self.batch_sample_conditions = torch.tensor_split(
            self.tensor_conditions, batch_num)

    def __iter__(self):
        """
        Return an iterator over the points. Any element of the iterator is a
        dictionary with the following keys:
            - ``pts``: The input sample points. It is a LabelTensor with the
                shape ``(batch_size, input_dimension)``.
            - ``output``: The output sample points. This key is present only
                if data conditions are present. It is a LabelTensor with the
                shape ``(batch_size, output_dimension)``.
            - ``condition``: The integer condition indeces. It is a tensor
                with the shape ``(batch_size, )`` of type ``torch.int64`` and
                indicates for any ``pts`` the corresponding problem condition.

        :return: An iterator over the points.
        :rtype: iter
        """
        #for i in self.random_idx:
        for i in range(len(self.batch_list)):
            type_, idx_ = self.batch_list[i]

            if type_ == 'sample':
                d = {
                    'pts': self.batch_sample_pts[idx_].requires_grad_(True),
                    'condition': self.batch_sample_conditions[idx_],
                }
            else:
                d = {
                    'pts': self.batch_input_pts[idx_].requires_grad_(True),
                    'output': self.batch_output_pts[idx_],
                    'condition': self.batch_data_conditions[idx_],
                }
            yield d


import torch
from torch.nn.parameter import Parameter


class AdaptiveSoftplus(torch.nn.Module):
    '''
    Implementation of soft exponential activation.
    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input
    Parameters:
        - alpha - trainable parameter
    References:
        - See related paper:
        https://arxiv.org/pdf/1602.01321.pdf
    Examples:
        >>> a1 = soft_exponential(256)
        >>> x = torch.randn(256)
        >>> x = a1(x)
    '''

    def __init__(self):
        '''
        Initialization.
        INPUT:
            - in_features: shape of the input
            - aplha: trainable parameter
            aplha is initialized with zero value by default
        '''
        super().__init__()

        self.soft = torch.nn.Softplus()

        self.scale = Parameter(torch.rand(1))
        self.scale.requiresGrad = True  # set requiresGrad to true!

    def forward(self, x):
        '''
        Forward pass of the function.
        Applies the function to the input elementwise.
        '''
        #x += self.translate
        return self.soft(x) * self.scale



""" Condition module. """
from .label_tensor import LabelTensor
from .geometry import Location
from .equation.equation import Equation


def dummy(a):
    """Dummy function for testing purposes."""
    return None


class Condition:
    """
    The class ``Condition`` is used to represent the constraints (physical
    equations, boundary conditions, etc.) that should be satisfied in the
    problem at hand. Condition objects are used to formulate the PINA :obj:`pina.problem.abstract_problem.AbstractProblem` object.
    Conditions can be specified in three ways:

        1. By specifying the input and output points of the condition; in such a
        case, the model is trained to produce the output points given the input
        points.

        2. By specifying the location and the equation of the condition; in such
        a case, the model is trained to minimize the equation residual by
        evaluating it at some samples of the location.

        3. By specifying the input points and the equation of the condition; in
        such a case, the model is trained to minimize the equation residual by
        evaluating it at the passed input points.

    Example::

    >>> example_domain = Span({'x': [0, 1], 'y': [0, 1]})
    >>> def example_dirichlet(input_, output_):
    >>>     value = 0.0
    >>>     return output_.extract(['u']) - value
    >>> example_input_pts = LabelTensor(
    >>>     torch.tensor([[0, 0, 0]]), ['x', 'y', 'z'])
    >>> example_output_pts = LabelTensor(torch.tensor([[1, 2]]), ['a', 'b'])
    >>>
    >>> Condition(
    >>>     input_points=example_input_pts,
    >>>     output_points=example_output_pts)
    >>> Condition(
    >>>     location=example_domain,
    >>>     equation=example_dirichlet)
    >>> Condition(
    >>>     input_points=example_input_pts,
    >>>     equation=example_dirichlet)

    """

    __slots__ = [
        'input_points', 'output_points', 'location', 'equation', 'data_weight'
    ]

    def _dictvalue_isinstance(self, dict_, key_, class_):
        """Check if the value of a dictionary corresponding to `key` is an instance of `class_`."""
        if key_ not in dict_.keys():
            return True

        return isinstance(dict_[key_], class_)

    def __init__(self, *args, **kwargs):
        """
        Constructor for the `Condition` class.
        """
        self.data_weight = kwargs.pop('data_weight', 1.0)

        if len(args) != 0:
            raise ValueError(
                'Condition takes only the following keyword arguments: {`input_points`, `output_points`, `location`, `function`, `data_weight`}.'
            )

        if (sorted(kwargs.keys()) != sorted(['input_points', 'output_points'])
                and sorted(kwargs.keys()) != sorted(['location', 'equation'])
                and sorted(kwargs.keys()) != sorted(
                    ['input_points', 'equation'])):
            raise ValueError(f'Invalid keyword arguments {kwargs.keys()}.')

        if not self._dictvalue_isinstance(kwargs, 'input_points', LabelTensor):
            raise TypeError('`input_points` must be a torch.Tensor.')
        if not self._dictvalue_isinstance(kwargs, 'output_points', LabelTensor):
            raise TypeError('`output_points` must be a torch.Tensor.')
        if not self._dictvalue_isinstance(kwargs, 'location', Location):
            raise TypeError('`location` must be a Location.')
        if not self._dictvalue_isinstance(kwargs, 'equation', Equation):
            raise TypeError('`equation` must be a Equation.')

        for key, value in kwargs.items():
            setattr(self, key, value)



""" Module """
from .equation import Equation
from ..operators import grad, div, laplacian


class FixedValue(Equation):

    def __init__(self, value, components=None):
        """
        Fixed Value Equation class. This class can be
        used to enforced a fixed value for a specific
        condition, e.g. Dirichlet Boundary conditions.

        :param float value: Value to be mantained fixed.
        :param list(str) components: the name of the output
            variables to calculate the gradient for. It should
            be a subset of the output labels. If ``None``,
            all the output variables are considered.
            Default is ``None``.
        """

        def equation(input_, output_):
            if components is None:
                return output_ - value
            return output_.extract(components) - value

        super().__init__(equation)


class FixedGradient(Equation):

    def __init__(self, value, components=None, d=None):
        """
        Fixed Gradient Equation class. This class can be
        used to enforced a fixed gradient for a specific
        condition.

        :param float value: Value to be mantained fixed.
        :param list(str) components: the name of the output
            variables to calculate the gradient for. It should
            be a subset of the output labels. If ``None``,
            all the output variables are considered.
            Default is ``None``.
        :param list(str) d: the name of the input variables on
            which the gradient is calculated. d should be a subset
            of the input labels. If ``None``, all the input variables
            are considered. Default is ``None``.
        """

        def equation(input_, output_):
            return grad(output_, input_, components=components, d=d) - value

        super().__init__(equation)


class FixedFlux(Equation):

    def __init__(self, value, components=None, d=None):
        """
        Fixed Flux Equation class. This class can be
        used to enforced a fixed flux for a specific
        condition.

        :param float value: Value to be mantained fixed.
        :param list(str) components: the name of the output
            variables to calculate the flux for. It should
            be a subset of the output labels. If ``None``,
            all the output variables are considered.
            Default is ``None``.
        :param list(str) d: the name of the input variables on
            which the flux is calculated. d should be a subset
            of the input labels. If ``None``, all the input variables
            are considered. Default is ``None``.
        """

        def equation(input_, output_):
            return div(output_, input_, components=components, d=d) - value

        super().__init__(equation)


class Laplace(Equation):

    def __init__(self, components=None, d=None):
        """
        Laplace Equation class. This class can be
        used to enforced a Laplace equation for a specific
        condition (force term set to zero).

        :param list(str) components: the name of the output
            variables to calculate the flux for. It should
            be a subset of the output labels. If ``None``,
            all the output variables are considered.
            Default is ``None``.
        :param list(str) d: the name of the input variables on
            which the flux is calculated. d should be a subset
            of the input labels. If ``None``, all the input variables
            are considered. Default is ``None``.
        """

        def equation(input_, output_):
            return laplacian(output_, input_, components=components, d=d)

        super().__init__(equation)



"""Module for Multi FeedForward model"""
import torch

from .feed_forward import FeedForward


class MultiFeedForward(torch.nn.Module):
    """
    The PINA implementation of MultiFeedForward network. 

    This model allows to create a network with multiple FeedForward combined
    together. The user has to define the `forward` method choosing how to
    combine the different FeedForward networks.

    :param dict ffn_dict: dictionary of FeedForward networks.
    """

    def __init__(self, ffn_dict):
        super().__init__()

        if not isinstance(ffn_dict, dict):
            raise TypeError

        for name, constructor_args in ffn_dict.items():
            setattr(self, name, FeedForward(**constructor_args))



"""Module for Union class. """

import torch
from .operation_interface import OperationInterface
from ..label_tensor import LabelTensor
import random


class Union(OperationInterface):

    def __init__(self, geometries):
        r"""
        PINA implementation of Unions of Domains.
        Given two sets :math:`A` and :math:`B` then the
        domain difference is defined as:

        .. math::
            A \cup B = \{x \mid x \in A \lor x \in B\},

        with :math:`x` a point in :math:`\mathbb{R}^N` and :math:`N`
        the dimension of the geometry space.

        :param list geometries: A list of geometries from ``pina.geometry`` 
            such as ``EllipsoidDomain`` or ``CartesianDomain``.

        :Example:
            >>> # Create two ellipsoid domains
            >>> ellipsoid1 = EllipsoidDomain({'x': [-1, 1], 'y': [-1, 1]})
            >>> ellipsoid2 = EllipsoidDomain({'x': [0, 2], 'y': [0, 2]})
            >>> # Create a union of the ellipsoid domains
            >>> union = GeometryUnion([ellipsoid1, ellipsoid2])

        """
        super().__init__(geometries)

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the ``Union`` domain.

        :param point: Point to be checked.
        :type point: LabelTensor
        :param check_border: Check if the point is also on the frontier
            of the ellipsoid, default ``False``.
        :type check_border: bool
        :return: Returning ``True`` if the point is inside, ``False`` otherwise.
        :rtype: bool
        """
        for geometry in self.geometries:
            if geometry.is_inside(point, check_border):
                return True
        return False

    def sample(self, n, mode='random', variables='all'):
        """
        Sample routine for ``Union`` domain.

        :param int n: Number of points to sample in the shape.
        :param str mode: Mode for sampling, defaults to ``random``. Available modes include: ``random``.
        :param variables: Variables to be sampled, defaults to ``all``.
        :type variables: str | list[str]
        :return: Returns ``LabelTensor`` of n sampled points.
        :rtype: LabelTensor

        :Example:
            >>> # Create two ellipsoid domains
            >>> cartesian1 = CartesianDomain({'x': [0, 2], 'y': [0, 2]})
            >>> cartesian2 = CartesianDomain({'x': [1, 3], 'y': [1, 3]})
            >>> # Create a union of the ellipsoid domains
            >>> union = Union([cartesian1, cartesian2])
            >>> # Sample
            >>> union.sample(n=5)
                LabelTensor([[1.2128, 2.1991],
                            [1.3530, 2.4317],
                            [2.2562, 1.6605],
                            [0.8451, 1.9878],
                            [1.8623, 0.7102]])
            >>> len(union.sample(n=5)
                5
        """
        sampled_points = []

        # calculate the number of points to sample for each geometry and the remainder
        remainder = n % len(self.geometries)
        num_points = n // len(self.geometries)

        # sample the points
        # NB. geometries as shuffled since if we sample
        # multiple times just one point, we would end
        # up sampling only from the first geometry.
        iter_ = random.sample(self.geometries, len(self.geometries))
        for i, geometry in enumerate(iter_):
            # int(i < remainder) is one only if we have a remainder
            # different than zero. Notice that len(geometries) is
            # always smaller than remaider.
            sampled_points.append(
                geometry.sample(num_points + int(i < remainder), mode,
                                variables))
            # in case number of sampled points is smaller than the number of geometries
            if len(sampled_points) >= n:
                break

        return LabelTensor(torch.cat(sampled_points), labels=self.variables)



import torch
import torch.nn as nn
from ...utils import check_consistency


class ResidualBlock(nn.Module):
    """Residual block base class. Implementation of a residual block.

    .. seealso::

        **Original reference**: He, Kaiming, et al.
        *Deep residual learning for image recognition.*
        Proceedings of the IEEE conference on computer vision
        and pattern recognition. 2016..
        DOI: `<https://arxiv.org/pdf/1512.03385.pdf>`_.

    """

    def __init__(self,
                 input_dim,
                 output_dim,
                 hidden_dim,
                 spectral_norm=False,
                 activation=torch.nn.ReLU()):
        """
        Initializes the ResidualBlock module.

        :param int input_dim: Dimension of the input to pass to the
            feedforward linear layer.
        :param int output_dim: Dimension of the output from the 
            residual layer.
        :param int hidden_dim: Hidden dimension for mapping the input
            (first block).
        :param bool spectral_norm: Apply spectral normalization to feedforward
            layers, defaults to False.
        :param torch.nn.Module activation: Cctivation function after first block.

        """
        super().__init__()
        # check consistency
        check_consistency(spectral_norm, bool)
        check_consistency(input_dim, int)
        check_consistency(output_dim, int)
        check_consistency(hidden_dim, int)
        check_consistency(activation, torch.nn.Module)

        # assign variables
        self._spectral_norm = spectral_norm
        self._input_dim = input_dim
        self._output_dim = output_dim
        self._hidden_dim = hidden_dim
        self._activation = activation

        # create layers
        self._l1 = self._spect_norm(nn.Linear(input_dim, hidden_dim))
        self._l2 = self._spect_norm(nn.Linear(hidden_dim, output_dim))
        self._l3 = self._spect_norm(nn.Linear(input_dim, output_dim))

    def forward(self, x):
        """Forward pass for residual block layer.

        :param torch.Tensor x: Input tensor for the residual layer.
        :return: Output tensor for the residual layer.
        :rtype: torch.Tensor
        """
        y = self._activation(self._l1(x))
        y = self._l2(y)
        x = self._l3(x)
        return y + x

    def _spect_norm(self, x):
        """Perform spectral norm on the layers.

        :param x: A torch.nn.Module Linear layer
        :type x: torch.nn.Module
        :return: The spectral norm of the layer
        :rtype: torch.nn.Module
        """
        return nn.utils.spectral_norm(x) if self._spectral_norm else x


import torch
import torch.nn as nn

class EnhancedLinear(torch.nn.Module):
    """
    A wrapper class for enhancing a linear layer with activation and/or dropout.

    :param layer: The linear layer to be enhanced.
    :type layer: torch.nn.Module
    :param activation: The activation function to be applied after the linear layer.
    :type activation: torch.nn.Module
    :param dropout: The dropout probability to be applied after the activation (if provided).
    :type dropout: float

    :Example:

    >>> linear_layer = torch.nn.Linear(10, 20)
    >>> activation = torch.nn.ReLU()
    >>> dropout_prob = 0.5
    >>> enhanced_linear = EnhancedLinear(linear_layer, activation, dropout_prob)
    """

    def __init__(self, layer, activation=None, dropout=None):
        """
        Initializes the EnhancedLinear module.

        :param layer: The linear layer to be enhanced.
        :type layer: torch.nn.Module
        :param activation: The activation function to be applied after the linear layer.
        :type activation: torch.nn.Module
        :param dropout: The dropout probability to be applied after the activation (if provided).
        :type dropout: float
        """
        super().__init__()

        # check consistency
        check_consistency(layer, nn.Module)
        if activation is not None:
            check_consistency(activation, nn.Module)
        if dropout is not None:
            check_consistency(dropout, float)

        # assign forward
        if (dropout is None) and (activation is None):
            self._model = torch.nn.Sequential(layer)

        elif (dropout is None) and (activation is not None):
            self._model = torch.nn.Sequential(layer, activation)

        elif (dropout is not None) and (activation is None):
            self._model = torch.nn.Sequential(layer, self._drop(dropout))

        elif (dropout is not None) and (activation is not None):
            self._model = torch.nn.Sequential(layer, activation,
                                              self._drop(dropout))

    def forward(self, x):
        """
        Forward pass through the enhanced linear module.

        :param x: Input tensor.
        :type x: torch.Tensor

        :return: Output tensor after passing through the enhanced linear module.
        :rtype: torch.Tensor
        """
        return self._model(x)

    def _drop(self, p):
        """
        Applies dropout with probability p.

        :param p: Dropout probability.
        :type p: float

        :return: Dropout layer with the specified probability.
        :rtype: torch.nn.Dropout
        """
        return torch.nn.Dropout(p)



"""Module for Continuous Convolution class"""
from .convolution import BaseContinuousConv
from .utils_convolution import check_point, map_points_
from .integral import Integral
import torch


class ContinuousConvBlock(BaseContinuousConv):
    """
    Implementation of Continuous Convolutional operator.

    The algorithm expects input to be in the form:
    :math:`[B, N_{in}, N,  D]`
    where :math:`B` is the batch_size, :math:`N_{in}` is the number of input
    fields, :math:`N` the number of points in the mesh, :math:`D` the dimension
    of the problem. In particular:

    *   :math:`D` is the number of spatial variables + 1. The last column must
        contain the field value. For example for 2D problems :math:`D=3` and
        the tensor will be something like ``[first coordinate, second
        coordinate, field value]``.
    *   :math:`N_{in}` represents the number of vectorial function presented.
        For example a vectorial function :math:`f = [f_1, f_2]` will have
        :math:`N_{in}=2`.

    .. seealso::

        **Original reference**: Coscia, D., Meneghetti, L., Demo, N. et al.
        *A continuous convolutional trainable filter for modelling unstructured data*.
        Comput Mech 72, 253–265 (2023). DOI `<https://doi.org/10.1007/s00466-023-02291-1>`_

    """

    def __init__(self,
                 input_numb_field,
                 output_numb_field,
                 filter_dim,
                 stride,
                 model=None,
                 optimize=False,
                 no_overlap=False):
        """
        :param input_numb_field: Number of fields :math:`N_{in}` in the input.
        :type input_numb_field: int
        :param output_numb_field: Number of fields :math:`N_{out}`  in the output.
        :type output_numb_field: int
        :param filter_dim: Dimension of the filter.
        :type filter_dim: tuple(int) | list(int)
        :param stride: Stride for the filter.
        :type stride: dict
        :param model: Neural network for inner parametrization,
            defaults to ``None``. If None, a default multilayer perceptron
            of width three and size twenty with ReLU activation is used.
        :type model: torch.nn.Module
        :param optimize: Flag for performing optimization on the continuous
            filter, defaults to False. The flag `optimize=True` should be
            used only when the scatter datapoints are fixed through the
            training. If torch model is in ``.eval()`` mode, the flag is
            automatically set to False always.
        :type optimize: bool
        :param no_overlap: Flag for performing optimization on the transpose
            continuous filter, defaults to False. The flag set to `True` should
            be used only when the filter positions do not overlap for different
            strides. RuntimeError will raise in case of non-compatible strides.
        :type no_overlap: bool

        .. note::
            Using `optimize=True` the filter can be use either in `forward`
            or in `transpose` mode, not both. If `optimize=False` the same
            filter can be used for both `transpose` and `forward` modes.

        :Example:
            >>> class MLP(torch.nn.Module):
                    def __init__(self) -> None:
                        super().__init__()
                        self. model = torch.nn.Sequential(
                                                        torch.nn.Linear(2, 8),
                                                        torch.nn.ReLU(),
                                                        torch.nn.Linear(8, 8),
                                                        torch.nn.ReLU(),
                                                        torch.nn.Linear(8, 1))
                    def forward(self, x):
                        return self.model(x)
            >>> dim = [3, 3]
            >>> stride = {"domain": [10, 10],
                          "start": [0, 0],
                          "jumps": [3, 3],
                          "direction": [1, 1.]}
            >>> conv = ContinuousConv2D(1, 2, dim, stride, MLP)
            >>> conv
                ContinuousConv2D(
                (_net): ModuleList(
                    (0): MLP(
                    (model): Sequential(
                        (0): Linear(in_features=2, out_features=8, bias=True)
                        (1): ReLU()
                        (2): Linear(in_features=8, out_features=8, bias=True)
                        (3): ReLU()
                        (4): Linear(in_features=8, out_features=1, bias=True)
                    )
                    )
                    (1): MLP(
                    (model): Sequential(
                        (0): Linear(in_features=2, out_features=8, bias=True)
                        (1): ReLU()
                        (2): Linear(in_features=8, out_features=8, bias=True)
                        (3): ReLU()
                        (4): Linear(in_features=8, out_features=1, bias=True)
                    )
                    )
                )
                )
        """

        super().__init__(input_numb_field=input_numb_field,
                         output_numb_field=output_numb_field,
                         filter_dim=filter_dim,
                         stride=stride,
                         model=model,
                         optimize=optimize,
                         no_overlap=no_overlap)

        # integral routine
        self._integral = Integral('discrete')

        # create the network
        self._net = self._spawn_networks(model)

        # stride for continuous convolution overridden
        self._stride = self._stride._stride_discrete

    def _spawn_networks(self, model):
        """
        Private method to create a collection of kernels

        :param model: A :class:`torch.nn.Module` model in form of Object class.
        :type model: torch.nn.Module
        :return: List of :class:`torch.nn.Module` models.
        :rtype: torch.nn.ModuleList

        """
        nets = []
        if self._net is None:
            for _ in range(self._input_numb_field * self._output_numb_field):
                tmp = ContinuousConvBlock.DefaultKernel(len(self._dim), 1)
                nets.append(tmp)
        else:
            if not isinstance(model, object):
                raise ValueError("Expected a python class inheriting"
                                 " from torch.nn.Module")

            for _ in range(self._input_numb_field * self._output_numb_field):
                tmp = model()
                if not isinstance(tmp, torch.nn.Module):
                    raise ValueError("The python class must be inherited from"
                                     " torch.nn.Module. See the docstring for"
                                     " an example.")
                nets.append(tmp)

        return torch.nn.ModuleList(nets)

    def _extract_mapped_points(self, batch_idx, index, x):
        """
        Priviate method to extract mapped points in the filter

        :param x: Input tensor of shape ``[channel, N, dim]``
        :type x: torch.Tensor
        :return: Mapped points and indeces for each channel,
        :rtype: torch.Tensor, list

        """
        mapped_points = []
        indeces_channels = []

        for stride_idx, current_stride in enumerate(self._stride):

            # indeces of points falling into filter range
            indeces = index[stride_idx][batch_idx]

            # how many points for each channel fall into the filter?
            numb_points_insiede = torch.sum(indeces, dim=-1).tolist()

            # extracting points for each channel
            # shape: [sum(numb_points_insiede), filter_dim + 1]
            point_stride = x[indeces]

            # mapping points in filter domain
            map_points_(point_stride[..., :-1], current_stride)

            # extracting points for each channel
            point_stride_channel = point_stride.split(numb_points_insiede)

            # appending in list for later use
            mapped_points.append(point_stride_channel)
            indeces_channels.append(numb_points_insiede)

        # stacking input for passing to neural net
        mapping = map(torch.cat, zip(*mapped_points))
        stacked_input = tuple(mapping)
        indeces_channels = tuple(zip(*indeces_channels))

        return stacked_input, indeces_channels

    def _find_index(self, X):
        """
        Private method to extract indeces for convolution.

        :param X: Input tensor, as in ContinuousConvBlock ``__init__``.
        :type X: torch.Tensor

        """
        # append the index for each stride
        index = []
        for _, current_stride in enumerate(self._stride):

            tmp = check_point(X, current_stride, self._dim)
            index.append(tmp)

        # storing the index
        self._index = index

    def _make_grid_forward(self, X):
        """
        Private method to create forward convolution grid.

        :param X: Input tensor, as in ContinuousConvBlock docstring.
        :type X: torch.Tensor

        """
        # filter dimension + number of points in output grid
        filter_dim = len(self._dim)
        number_points = len(self._stride)

        # initialize the grid
        grid = torch.zeros(size=(X.shape[0], self._output_numb_field,
                                 number_points, filter_dim + 1),
                           device=X.device,
                           dtype=X.dtype)
        grid[..., :-1] = (self._stride + self._dim * 0.5)

        # saving the grid
        self._grid = grid.detach()

    def _make_grid_transpose(self, X):
        """
        Private method to create transpose convolution grid.

        :param X: Input tensor, as in ContinuousConvBlock docstring.
        :type X: torch.Tensor


        """
        # initialize to all zeros
        tmp = torch.zeros_like(X)
        tmp[..., :-1] = X[..., :-1]

        # save on tmp
        self._grid_transpose = tmp

    def _make_grid(self, X, type):
        """
        Private method to create convolution grid.

        :param X: Input tensor, as in ContinuousConvBlock docstring.
        :type X: torch.Tensor
        :param type: Type of convolution, ``['forward', 'inverse']`` the
            possibilities.
        :type type: str

        """
        # choose the type of convolution
        if type == 'forward':
            return self._make_grid_forward(X)
        elif type == 'inverse':
            self._make_grid_transpose(X)
        else:
            raise TypeError

    def _initialize_convolution(self, X, type='forward'):
        """
        Private method to intialize the convolution.
        The convolution is initialized by setting a grid and
        calculate the index for finding the points inside the
        filter.

        :param X: Input tensor, as in ContinuousConvBlock docstring.
        :type X: torch.Tensor
        :param str type: type of convolution, ``['forward', 'inverse'] ``the
            possibilities.
        """

        # variable for the convolution
        self._make_grid(X, type)

        # calculate the index
        self._find_index(X)

    def forward(self, X):
        """
        Forward pass in the convolutional layer.

        :param x: Input data for the convolution :math:`[B, N_{in}, N,  D]`.
        :type x: torch.Tensor
        :return: Convolution output :math:`[B, N_{out}, N,  D]`.
        :rtype: torch.Tensor
        """

        # initialize convolution
        if self.training:  # we choose what to do based on optimization
            self._choose_initialization(X, type='forward')

        else:  # we always initialize on testing
            self._initialize_convolution(X, 'forward')

        # create convolutional array
        conv = self._grid.clone().detach()

        # total number of fields
        tot_dim = self._output_numb_field * self._input_numb_field

        for batch_idx, x in enumerate(X):

            # extract mapped points
            stacked_input, indeces_channels = self._extract_mapped_points(
                batch_idx, self._index, x)

            # compute the convolution

            # storing intermidiate results for each channel convolution
            res_tmp = []
            # for each field
            for idx_conv in range(tot_dim):
                # index for each input field
                idx = idx_conv % self._input_numb_field
                # extract input for each channel
                single_channel_input = stacked_input[idx]
                # extract filter
                net = self._net[idx_conv]
                # calculate filter value
                staked_output = net(single_channel_input[..., :-1])
                # perform integral for all strides in one field
                integral = self._integral(staked_output,
                                          single_channel_input[..., -1],
                                          indeces_channels[idx])
                res_tmp.append(integral)

            # stacking integral results
            res_tmp = torch.stack(res_tmp)

            # sum filters (for each input fields) in groups
            # for different ouput fields
            conv[batch_idx, ...,
                 -1] = res_tmp.reshape(self._output_numb_field,
                                       self._input_numb_field, -1).sum(1)
        return conv

    def transpose_no_overlap(self, integrals, X):
        """
        Transpose pass in the layer for no-overlapping filters

        :param integrals: Weights for the transpose convolution. Shape
            :math:`[B, N_{in}, N]`
            where B is the batch_size, :math`N_{in}` is the number of input
            fields, :math:`N` the number of points in the mesh, D the dimension
            of the problem.
        :type integral: torch.tensor
        :param X: Input data. Expect tensor of shape
            :math:`[B, N_{in}, M,  D]` where :math:`B` is the batch_size,
            :math`N_{in}`is the number of input fields, :math:`M` the number of points
            in the mesh, :math:`D` the dimension of the problem.
        :type X: torch.Tensor
        :return: Feed forward transpose convolution. Tensor of shape
            :math:`[B, N_{out}, M,  D]` where :math:`B` is the batch_size,
            :math`N_{out}`is the number of input fields, :math:`M` the number of points
            in the mesh, :math:`D` the dimension of the problem.
        :rtype: torch.Tensor

        .. note::
            This function is automatically called when ``.transpose()``
            method is used and ``no_overlap=True``
        """

        # initialize convolution
        if self.training:  # we choose what to do based on optimization
            self._choose_initialization(X, type='inverse')

        else:  # we always initialize on testing
            self._initialize_convolution(X, 'inverse')

        # initialize grid
        X = self._grid_transpose.clone().detach()
        conv_transposed = self._grid_transpose.clone().detach()

        # total number of dim
        tot_dim = self._input_numb_field * self._output_numb_field

        for batch_idx, x in enumerate(X):

            # extract mapped points
            stacked_input, indeces_channels = self._extract_mapped_points(
                batch_idx, self._index, x)

            # compute the transpose convolution

            # total number of fields
            res_tmp = []

            # for each field
            for idx_conv in range(tot_dim):
                # index for each output field
                idx = idx_conv % self._output_numb_field
                # index for each input field
                idx_in = idx_conv % self._input_numb_field
                # extract input for each field
                single_channel_input = stacked_input[idx]
                rep_idx = torch.tensor(indeces_channels[idx])
                integral = integrals[batch_idx,
                                     idx_in, :].repeat_interleave(rep_idx)
                # extract filter
                net = self._net[idx_conv]
                # perform transpose convolution for all strides in one field
                staked_output = net(single_channel_input[..., :-1]).flatten()
                integral = staked_output * integral
                res_tmp.append(integral)

            # stacking integral results and sum
            # filters (for each input fields) in groups
            # for different output fields
            res_tmp = torch.stack(res_tmp).reshape(self._input_numb_field,
                                                   self._output_numb_field,
                                                   -1).sum(0)
            conv_transposed[batch_idx, ..., -1] = res_tmp

        return conv_transposed

    def transpose_overlap(self, integrals, X):
        """
        Transpose pass in the layer for overlapping filters

        :param integrals: Weights for the transpose convolution. Shape
            :math:`[B, N_{in}, N]`
            where B is the batch_size, :math`N_{in}` is the number of input
            fields, :math:`N` the number of points in the mesh, D the dimension
            of the problem.
        :type integral: torch.tensor
        :param X: Input data. Expect tensor of shape
            :math:`[B, N_{in}, M,  D]` where :math:`B` is the batch_size,
            :math`N_{in}`is the number of input fields, :math:`M` the number of points
            in the mesh, :math:`D` the dimension of the problem.
        :type X: torch.Tensor
        :return: Feed forward transpose convolution. Tensor of shape
            :math:`[B, N_{out}, M,  D]` where :math:`B` is the batch_size,
            :math`N_{out}`is the number of input fields, :math:`M` the number of points
            in the mesh, :math:`D` the dimension of the problem.
        :rtype: torch.Tensor

        .. note:: This function is automatically called when ``.transpose()``
            method is used and ``no_overlap=False``
        """

        # initialize convolution
        if self.training:  # we choose what to do based on optimization
            self._choose_initialization(X, type='inverse')

        else:  # we always initialize on testing
            self._initialize_convolution(X, 'inverse')

        # initialize grid
        X = self._grid_transpose.clone().detach()
        conv_transposed = self._grid_transpose.clone().detach()

        # list to iterate for calculating nn output
        tmp = [i for i in range(self._output_numb_field)]
        iterate_conv = [
            item for item in tmp for _ in range(self._input_numb_field)
        ]

        for batch_idx, x in enumerate(X):

            # accumulator for the convolution on different batches
            accumulator_batch = torch.zeros(
                size=(self._grid_transpose.shape[1],
                      self._grid_transpose.shape[2]),
                requires_grad=True,
                device=X.device,
                dtype=X.dtype).clone()

            for stride_idx, current_stride in enumerate(self._stride):
                # indeces of points falling into filter range
                indeces = self._index[stride_idx][batch_idx]

                # number of points for each channel
                numb_pts_channel = tuple(indeces.sum(dim=-1))

                # extracting points for each channel
                point_stride = x[indeces]

                # if no points to upsample we just skip
                if point_stride.nelement() == 0:
                    continue

                # mapping points in filter domain
                map_points_(point_stride[..., :-1], current_stride)

                # input points for kernels
                # we split for extracting number of points for each channel
                nn_input_pts = point_stride[..., :-1].split(numb_pts_channel)

                # accumulate partial convolution results for each field
                res_tmp = []

                # for each channel field compute transpose convolution
                for idx_conv, idx_channel_out in enumerate(iterate_conv):

                    # index for input channels
                    idx_channel_in = idx_conv % self._input_numb_field

                    # extract filter
                    net = self._net[idx_conv]

                    # calculate filter value
                    staked_output = net(nn_input_pts[idx_channel_out])

                    # perform integral for all strides in one field
                    integral = staked_output * integrals[batch_idx,
                                                         idx_channel_in,
                                                         stride_idx]
                    # append results
                    res_tmp.append(integral.flatten())

                # computing channel sum
                channel_sum = []
                start = 0
                for _ in range(self._output_numb_field):
                    tmp = res_tmp[start:start + self._input_numb_field]
                    tmp = torch.vstack(tmp).sum(dim=0)
                    channel_sum.append(tmp)
                    start += self._input_numb_field

                # accumulate the results
                accumulator_batch[indeces] += torch.hstack(channel_sum)

            # save results of accumulation for each batch
            conv_transposed[batch_idx, ..., -1] = accumulator_batch

        return conv_transposed



"""Module for Difference class."""

import torch
from .operation_interface import OperationInterface
from ..label_tensor import LabelTensor


class Difference(OperationInterface):

    def __init__(self, geometries):
        r"""
        PINA implementation of Difference of Domains.
        Given two sets :math:`A` and :math:`B` then the
        domain difference is defined as:

        .. math::
            A - B = \{x \mid x \in A \land x \not\in B\},

        with :math:`x` a point in :math:`\mathbb{R}^N` and :math:`N`
        the dimension of the geometry space.

        :param list geometries: A list of geometries from ``pina.geometry``
            such as ``EllipsoidDomain`` or ``CartesianDomain``. The first 
            geometry in the list is the geometry from which points are
            sampled. The rest of the geometries are the geometries that
            are excluded from the first geometry to find the difference.

        :Example:
            >>> # Create two ellipsoid domains
            >>> ellipsoid1 = EllipsoidDomain({'x': [-1, 1], 'y': [-1, 1]})
            >>> ellipsoid2 = EllipsoidDomain({'x': [0, 2], 'y': [0, 2]})
            >>> # Create a Difference of the ellipsoid domains
            >>> difference = Difference([ellipsoid1, ellipsoid2])
        """
        super().__init__(geometries)

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the ``Difference`` domain.

        :param point: Point to be checked.
        :type point: torch.Tensor   
        :param bool check_border: If ``True``, the border is considered inside.
        :return: ``True`` if the point is inside the Exclusion domain, ``False`` otherwise.
        :rtype: bool
        """
        for geometry in self.geometries[1:]:
            if geometry.is_inside(point):
                return False
        return self.geometries[0].is_inside(point, check_border)

    def sample(self, n, mode='random', variables='all'):
        """
        Sample routine for ``Difference`` domain.

        :param int n: Number of points to sample in the shape.
        :param str mode: Mode for sampling, defaults to ``random``. Available modes include: ``random``.
        :param variables: Variables to be sampled, defaults to ``all``.
        :type variables: str | list[str]
        :return: Returns ``LabelTensor`` of n sampled points.
        :rtype: LabelTensor

        :Example:
            >>> # Create two Cartesian domains
            >>> cartesian1 = CartesianDomain({'x': [0, 2], 'y': [0, 2]})
            >>> cartesian2 = CartesianDomain({'x': [1, 3], 'y': [1, 3]})
            >>> # Create a Difference of the ellipsoid domains
            >>> difference = Difference([cartesian1, cartesian2])
            >>> # Sampling
            >>> difference.sample(n=5)
                LabelTensor([[0.8400, 0.9179],
                            [0.9154, 0.5769],
                            [1.7403, 0.4835],
                            [0.9545, 1.2851],
                            [1.3726, 0.9831]])
            >>> len(difference.sample(n=5)
                5

        """
        if mode != 'random':
            raise NotImplementedError(
                f'{mode} is not a valid mode for sampling.')

        sampled = []

        # sample the points
        while len(sampled) < n:
            # get sample point from first geometry
            point = self.geometries[0].sample(1, mode, variables)
            is_inside = False

            # check if point is inside any other geometry
            for geometry in self.geometries[1:]:
                # if point is inside any other geometry, break
                if geometry.is_inside(point):
                    is_inside = True
                    break
            # if point is not inside any other geometry, add to sampled
            if not is_inside:
                sampled.append(point)

        return LabelTensor(torch.cat(sampled), labels=self.variables)



""" Module for LabelTensor """
from typing import Any
import torch
from torch import Tensor


class LabelTensor(torch.Tensor):
    """Torch tensor with a label for any column."""

    @staticmethod
    def __new__(cls, x, labels, *args, **kwargs):
        return super().__new__(cls, x, *args, **kwargs)

    def __init__(self, x, labels):
        '''
        Construct a `LabelTensor` by passing a tensor and a list of column
        labels. Such labels uniquely identify the columns of the tensor,
        allowing for an easier manipulation.

        :param torch.Tensor x: The data tensor.
        :param labels: The labels of the columns.
        :type labels: str | list(str) | tuple(str)

        :Example:
            >>> from pina import LabelTensor
            >>> tensor = LabelTensor(torch.rand((2000, 3)), ['a', 'b', 'c'])
            >>> tensor
            tensor([[6.7116e-02, 4.8892e-01, 8.9452e-01],
                [9.2392e-01, 8.2065e-01, 4.1986e-04],
                [8.9266e-01, 5.5446e-01, 6.3500e-01],
                ...,
                [5.8194e-01, 9.4268e-01, 4.1841e-01],
                [1.0246e-01, 9.5179e-01, 3.7043e-02],
                [9.6150e-01, 8.0656e-01, 8.3824e-01]])
            >>> tensor.extract('a')
            tensor([[0.0671],
                    [0.9239],
                    [0.8927],
                    ...,
                    [0.5819],
                    [0.1025],
                    [0.9615]])
            >>> tensor['a']
            tensor([[0.0671],
                    [0.9239],
                    [0.8927],
                    ...,
                    [0.5819],
                    [0.1025],
                    [0.9615]])
            >>> tensor.extract(['a', 'b'])
            tensor([[0.0671, 0.4889],
                    [0.9239, 0.8207],
                    [0.8927, 0.5545],
                    ...,
                    [0.5819, 0.9427],
                    [0.1025, 0.9518],
                    [0.9615, 0.8066]])
            >>> tensor.extract(['b', 'a'])
            tensor([[0.4889, 0.0671],
                    [0.8207, 0.9239],
                    [0.5545, 0.8927],
                    ...,
                    [0.9427, 0.5819],
                    [0.9518, 0.1025],
                    [0.8066, 0.9615]])
        '''
        if x.ndim == 1:
            x = x.reshape(-1, 1)

        if isinstance(labels, str):
            labels = [labels]

        if len(labels) != x.shape[-1]:
            raise ValueError('the tensor has not the same number of columns of '
                             'the passed labels.')
        self._labels = labels
    
    @property
    def labels(self):
        """Property decorator for labels

        :return: labels of self
        :rtype: list
        """
        return self._labels

    @labels.setter
    def labels(self, labels):
        if len(labels) != self.shape[self.ndim - 1]:  # small check
            raise ValueError('The tensor has not the same number of columns of '
                             'the passed labels.')

        self._labels = labels  # assign the label

    @staticmethod
    def vstack(label_tensors):
        """
        Stack tensors vertically. For more details, see
        :meth:`torch.vstack`.

        :param list(LabelTensor) label_tensors: the tensors to stack. They need
            to have equal labels.
        :return: the stacked tensor
        :rtype: LabelTensor
        """
        if len(label_tensors) == 0:
            return []

        all_labels = [label for lt in label_tensors for label in lt.labels]
        if set(all_labels) != set(label_tensors[0].labels):
            raise RuntimeError('The tensors to stack have different labels')

        labels = label_tensors[0].labels
        tensors = [lt.extract(labels) for lt in label_tensors]
        return LabelTensor(torch.vstack(tensors), labels)

    def clone(self, *args, **kwargs):
        """
        Clone the LabelTensor. For more details, see
        :meth:`torch.Tensor.clone`.

        :return: A copy of the tensor.
        :rtype: LabelTensor
        """
        # # used before merging 
        # try:
        #     out = LabelTensor(super().clone(*args, **kwargs), self.labels)
        # except:
        #     out = super().clone(*args, **kwargs)
        out = LabelTensor(super().clone(*args, **kwargs), self.labels)
        return out

    def to(self, *args, **kwargs):
        """
        Performs Tensor dtype and/or device conversion. For more details, see
        :meth:`torch.Tensor.to`.
        """
        tmp = super().to(*args, **kwargs)
        new = self.__class__.clone(self)
        new.data = tmp.data
        return new

    def select(self, *args, **kwargs):
        """
        Performs Tensor selection. For more details, see :meth:`torch.Tensor.select`.
        """
        tmp = super().select(*args, **kwargs)
        tmp._labels = self._labels
        return tmp

    def cuda(self, *args, **kwargs):
        """
        Send Tensor to cuda. For more details, see :meth:`torch.Tensor.cuda`.
        """
        tmp = super().cuda(*args, **kwargs)
        new = self.__class__.clone(self)
        new.data = tmp.data
        return tmp

    def cpu(self, *args, **kwargs):
        """
        Send Tensor to cpu. For more details, see :meth:`torch.Tensor.cpu`.
        """
        tmp = super().cpu(*args, **kwargs)
        new = self.__class__.clone(self)
        new.data = tmp.data
        return tmp

    def extract(self, label_to_extract):
        """
        Extract the subset of the original tensor by returning all the columns
        corresponding to the passed ``label_to_extract``.

        :param label_to_extract: The label(s) to extract.
        :type label_to_extract: str | list(str) | tuple(str)
        :raises TypeError: Labels are not ``str``.
        :raises ValueError: Label to extract is not in the labels ``list``.
        """

        if isinstance(label_to_extract, str):
            label_to_extract = [label_to_extract]
        elif isinstance(label_to_extract, (tuple, list)):  # TODO
            pass
        else:
            raise TypeError(
                '`label_to_extract` should be a str, or a str iterator')

        indeces = []
        for f in label_to_extract:
            try:
                indeces.append(self.labels.index(f))
            except ValueError:
                raise ValueError(f'`{f}` not in the labels list')

        new_data = super(Tensor, self.T).__getitem__(indeces).T
        new_labels = [self.labels[idx] for idx in indeces]

        extracted_tensor = new_data.as_subclass(LabelTensor)
        extracted_tensor.labels = new_labels

        return extracted_tensor

    def detach(self):
        detached = super().detach()
        if hasattr(self, '_labels'):
            detached._labels = self._labels
        return detached


    def requires_grad_(self, mode = True):
        lt = super().requires_grad_(mode)
        lt.labels = self.labels
        return lt

    def append(self, lt, mode='std'):
        """
        Return a copy of the merged tensors.

        :param LabelTensor lt: The tensor to merge.
        :param str mode: {'std', 'first', 'cross'}
        :return: The merged tensors.
        :rtype: LabelTensor
        """
        if set(self.labels).intersection(lt.labels):
            raise RuntimeError('The tensors to merge have common labels')

        new_labels = self.labels + lt.labels
        if mode == 'std':
            new_tensor = torch.cat((self, lt), dim=1)
        elif mode == 'first':
            raise NotImplementedError
        elif mode == 'cross':
            tensor1 = self
            tensor2 = lt
            n1 = tensor1.shape[0]
            n2 = tensor2.shape[0]

            tensor1 = LabelTensor(tensor1.repeat(n2, 1), labels=tensor1.labels)
            tensor2 = LabelTensor(tensor2.repeat_interleave(n1, dim=0),
                                  labels=tensor2.labels)
            new_tensor = torch.cat((tensor1, tensor2), dim=1)

        new_tensor = new_tensor.as_subclass(LabelTensor)
        new_tensor.labels = new_labels
        return new_tensor

    def __getitem__(self, index):
        """
        Return a copy of the selected tensor.
        """

        if isinstance(index, str) or (isinstance(index, (tuple, list))and all(isinstance(a, str) for a in index)):
            return self.extract(index)

        selected_lt = super(Tensor, self).__getitem__(index)
        
        try:
            len_index = len(index)
        except TypeError:
            len_index = 1
        
        if isinstance(index, int) or len_index == 1:
            if selected_lt.ndim == 1:
                selected_lt = selected_lt.reshape(1, -1)
            if hasattr(self, 'labels'):
                selected_lt.labels = self.labels
        elif len_index == 2:
            if selected_lt.ndim == 1:
                selected_lt = selected_lt.reshape(-1, 1)
            if hasattr(self, 'labels'):
                if isinstance(index[1], list):
                    selected_lt.labels = [self.labels[i] for i in index[1]]
                else:
                    selected_lt.labels = self.labels[index[1]]
        else:
            selected_lt.labels = self.labels
                    
        return selected_lt
    
    @property
    def tensor(self):
        return self.as_subclass(Tensor)

    def __len__(self) -> int:
        return super().__len__()

    def __str__(self):
        if hasattr(self, 'labels'):
            s = f'labels({str(self.labels)})\n'
        else:
            s = 'no labels\n'
        s += super().__str__()
        return s


"""Module for Base Continuous Convolution class."""
from abc import ABCMeta, abstractmethod
import torch
from .stride import Stride
from .utils_convolution import optimizing


class BaseContinuousConv(torch.nn.Module, metaclass=ABCMeta):
    """
    Abstract class
    """

    def __init__(self,
                 input_numb_field,
                 output_numb_field,
                 filter_dim,
                 stride,
                 model=None,
                 optimize=False,
                 no_overlap=False):
        """
        Base Class for Continuous Convolution.

        The algorithm expects input to be in the form:
        $$[B \times N_{in} \times N \times D]$$
        where $B$ is the batch_size, $N_{in}$ is the number of input
        fields, $N$ the number of points in the mesh, $D$ the dimension
        of the problem. In particular:
        * $D$ is the number of spatial variables + 1. The last column must
            contain the field value. For example for 2D problems $D=3$ and
            the tensor will be something like `[first coordinate, second
            coordinate, field value]`.
        * $N_{in}$ represents the number of vectorial function presented.
            For example a vectorial function $f = [f_1, f_2]$ will have
            $N_{in}=2$.

        :Note
            A 2-dimensional vectorial function $N_{in}=2$ of 3-dimensional
            input $D=3+1=4$ with 100 points input mesh and batch size of 8
            is represented as a tensor `[8, 2, 100, 4]`, where the columns
            `[:, 0, :, -1]` and `[:, 1, :, -1]` represent the first and
            second filed value respectively

        The algorithm returns a tensor of shape:
        $$[B \times N_{out} \times N' \times D]$$
        where $B$ is the batch_size, $N_{out}$ is the number of output
        fields, $N'$ the number of points in the mesh, $D$ the dimension
        of the problem.

        :param input_numb_field: number of fields in the input
        :type input_numb_field: int
        :param output_numb_field: number of fields in the output
        :type output_numb_field: int
        :param filter_dim: dimension of the filter
        :type filter_dim: tuple/ list
        :param stride: stride for the filter
        :type stride: dict
        :param model: neural network for inner parametrization,
            defaults to None.
        :type model: torch.nn.Module, optional
        :param optimize: flag for performing optimization on the continuous
            filter, defaults to False. The flag `optimize=True` should be
            used only when the scatter datapoints are fixed through the
            training. If torch model is in `.eval()` mode, the flag is
            automatically set to False always.
        :type optimize: bool, optional
        :param no_overlap: flag for performing optimization on the transpose
            continuous filter, defaults to False. The flag set to `True` should
            be used only when the filter positions do not overlap for different
            strides. RuntimeError will raise in case of non-compatible strides.
        :type no_overlap: bool, optional
        """
        super().__init__()

        if isinstance(input_numb_field, int):
            self._input_numb_field = input_numb_field
        else:
            raise ValueError('input_numb_field must be int.')

        if isinstance(output_numb_field, int):
            self._output_numb_field = output_numb_field
        else:
            raise ValueError('input_numb_field must be int.')

        if isinstance(filter_dim, (tuple, list)):
            vect = filter_dim
        else:
            raise ValueError('filter_dim must be tuple or list.')
        vect = torch.tensor(vect)
        self.register_buffer("_dim", vect, persistent=False)

        if isinstance(stride, dict):
            self._stride = Stride(stride)
        else:
            raise ValueError('stride must be dictionary.')

        self._net = model

        if isinstance(optimize, bool):
            self._optimize = optimize
        else:
            raise ValueError('optimize must be bool.')

        # choosing how to initialize based on optimization
        if self._optimize:
            # optimizing decorator ensure the function is called
            # just once
            self._choose_initialization = optimizing(
                self._initialize_convolution)
        else:
            self._choose_initialization = self._initialize_convolution

        if not isinstance(no_overlap, bool):
            raise ValueError('no_overlap must be bool.')

        if no_overlap:
            raise NotImplementedError
            self.transpose = self.transpose_no_overlap
        else:
            self.transpose = self.transpose_overlap

    class DefaultKernel(torch.nn.Module):

        def __init__(self, input_dim, output_dim):
            super().__init__()
            assert isinstance(input_dim, int)
            assert isinstance(output_dim, int)
            self._model = torch.nn.Sequential(torch.nn.Linear(input_dim, 20),
                                              torch.nn.ReLU(),
                                              torch.nn.Linear(20, 20),
                                              torch.nn.ReLU(),
                                              torch.nn.Linear(20, output_dim))

        def forward(self, x):
            return self._model(x)

    @property
    def net(self):
        return self._net

    @property
    def stride(self):
        return self._stride

    @property
    def filter_dim(self):
        return self._dim

    @property
    def input_numb_field(self):
        return self._input_numb_field

    @property
    def output_numb_field(self):
        return self._output_numb_field

    @property
    @abstractmethod
    def forward(self, X):
        pass

    @property
    @abstractmethod
    def transpose_overlap(self, X):
        pass

    @property
    @abstractmethod
    def transpose_no_overlap(self, X):
        pass

    @property
    @abstractmethod
    def _initialize_convolution(self, X, type):
        pass



'''PINA Callbacks Implementations'''

# from lightning.pytorch.callbacks import Callback
from pytorch_lightning.callbacks import Callback
import torch
from ..utils import check_consistency


class R3Refinement(Callback):


    def __init__(self, sample_every):
        """
        PINA Implementation of an R3 Refinement Callback.

        This callback implements the R3 (Retain-Resample-Release) routine for sampling new points based on adaptive search.
        The algorithm incrementally accumulates collocation points in regions of high PDE residuals, and releases those 
        with low residuals. Points are sampled uniformly in all regions where sampling is needed.

        .. seealso::

            Original Reference: Daw, Arka, et al. *Mitigating Propagation Failures in Physics-informed Neural Networks 
            using Retain-Resample-Release (R3) Sampling. (2023)*.
            DOI: `10.48550/arXiv.2207.02338
            <https://doi.org/10.48550/arXiv.2207.02338>`_

        :param int sample_every: Frequency for sampling.

        :raises ValueError: If `sample_every` is not an integer.

        Example:
            >>> r3_callback = R3Refinement(sample_every=5)
        """
        super().__init__()

        # sample every
        check_consistency(sample_every, int)
        self._sample_every = sample_every

    def _compute_residual(self, trainer):
        """
        Computes the residuals for a PINN object.

        :return: the total loss, and pointwise loss.
        :rtype: tuple
        """

        # extract the solver and device from trainer
        solver = trainer._model
        device = trainer._accelerator_connector._accelerator_flag

        # compute residual
        res_loss = {}
        tot_loss = []
        for location in self._sampling_locations:
            condition = solver.problem.conditions[location]
            pts = solver.problem.input_pts[location]
            # send points to correct device
            pts = pts.to(device)
            pts = pts.requires_grad_(True)
            pts.retain_grad()
            # PINN loss: equation evaluated only on locations where sampling is needed
            target = condition.equation.residual(pts, solver.forward(pts))
            res_loss[location] = torch.abs(target).as_subclass(torch.Tensor)
            tot_loss.append(torch.abs(target))

        return torch.vstack(tot_loss), res_loss

    def _r3_routine(self, trainer):
        """
        R3 refinement main routine.

        :param Trainer trainer: PINA Trainer.
        """
        # compute residual (all device possible)
        tot_loss, res_loss = self._compute_residual(trainer)
        tot_loss = tot_loss.as_subclass(torch.Tensor)

        # !!!!!! From now everything is performed on CPU !!!!!!

        # average loss
        avg = (tot_loss.mean()).to('cpu')

        # points to keep
        old_pts = {}
        tot_points = 0
        for location in self._sampling_locations:
            pts = trainer._model.problem.input_pts[location]
            labels = pts.labels
            pts = pts.cpu().detach()
            residuals = res_loss[location].cpu()
            mask = (residuals > avg).flatten()
            if any(mask): # if there are residuals greater than averge we append them
                pts = pts[mask] # TODO masking remove labels
                pts.labels = labels
                old_pts[location] = pts
                tot_points += len(pts)

        # extract new points to sample uniformally for each location
        n_points = (self._tot_pop_numb - tot_points) // len(
            self._sampling_locations)
        remainder = (self._tot_pop_numb - tot_points) % len(
            self._sampling_locations)
        n_uniform_points = [n_points] * len(self._sampling_locations)
        n_uniform_points[-1] += remainder

        # sample new points
        for numb_pts, loc in zip(n_uniform_points, self._sampling_locations):
            trainer._model.problem.discretise_domain(numb_pts,
                                                     'random',
                                                     locations=[loc])
        # adding previous population points
        trainer._model.problem.add_points(old_pts)

        # update dataloader
        trainer._create_or_update_loader()

    def on_train_start(self, trainer, _):
        """
        Callback function called at the start of training.

        This method extracts the locations for sampling from the problem conditions and calculates the total population.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer
        :param _: Placeholder argument (not used).

        :return: None
        :rtype: None
        """
        # extract locations for sampling
        problem = trainer._model.problem
        locations = []
        for condition_name in problem.conditions:
            condition = problem.conditions[condition_name]
            if hasattr(condition, 'location'):
                locations.append(condition_name)
        self._sampling_locations = locations

        # extract total population
        total_population = 0
        for location in self._sampling_locations:
            pts = trainer._model.problem.input_pts[location]
            total_population += len(pts)
        self._tot_pop_numb = total_population

    def on_train_epoch_end(self, trainer, __):
        """
        Callback function called at the end of each training epoch.

        This method triggers the R3 routine for refinement if the current epoch is a multiple of `_sample_every`.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer
        :param __: Placeholder argument (not used).

        :return: None
        :rtype: None
        """
        if trainer.current_epoch % self._sample_every == 0:
            self._r3_routine(trainer)



"""
Module for operators vectorize implementation. Differential operators are used to write any differential problem.
These operators are implemented to work on different accellerators: CPU, GPU, TPU or MPS.
All operators take as input a tensor onto which computing the operator, a tensor with respect
to which computing the operator, the name of the output variables to calculate the operator
for (in case of multidimensional functions), and the variables name on which the operator is calculated.
"""
import torch

from pina.label_tensor import LabelTensor


def grad(output_, input_, components=None, d=None):
    """
    Perform gradient operation. The operator works for vectorial and scalar
    functions, with multiple input coordinates.

    :param LabelTensor output_: the output tensor onto which computing the
        gradient.
    :param LabelTensor input_: the input tensor with respect to which computing
        the gradient.
    :param list(str) components: the name of the output variables to calculate
        the gradient for. It should be a subset of the output labels. If None,
        all the output variables are considered. Default is None.
    :param list(str) d: the name of the input variables on which the gradient is
        calculated. d should be a subset of the input labels. If None, all the
        input variables are considered. Default is None.

    :return: the gradient tensor.
    :rtype: LabelTensor
    """

    def grad_scalar_output(output_, input_, d):
        """
        Perform gradient operation for a scalar output.

        :param LabelTensor output_: the output tensor onto which computing the
            gradient. It has to be a column tensor.
        :param LabelTensor input_: the input tensor with respect to which
            computing the gradient.
        :param list(str) d: the name of the input variables on which the
            gradient is calculated. d should be a subset of the input labels. If
            None, all the input variables are considered. Default is None.

        :raises RuntimeError: a vectorial function is passed.
        :raises RuntimeError: missing derivative labels.
        :return: the gradient tensor.
        :rtype: LabelTensor
        """

        if len(output_.labels) != 1:
            raise RuntimeError('only scalar function can be differentiated')
        if not all([di in input_.labels for di in d]):
            raise RuntimeError('derivative labels missing from input tensor')

        output_fieldname = output_.labels[0]
        gradients = torch.autograd.grad(output_,
                                        input_,
                                        grad_outputs=torch.ones(
                                            output_.size(),
                                            dtype=output_.dtype,
                                            device=output_.device),
                                        create_graph=True,
                                        retain_graph=True,
                                        allow_unused=True)[0]

        gradients.labels = input_.labels
        gradients = gradients.extract(d)
        gradients.labels = [f'd{output_fieldname}d{i}' for i in d]

        return gradients

    if not isinstance(input_, LabelTensor):
        raise TypeError

    if d is None:
        d = input_.labels

    if components is None:
        components = output_.labels

    if output_.shape[1] == 1:  # scalar output ################################

        if components != output_.labels:
            raise RuntimeError
        gradients = grad_scalar_output(output_, input_, d)

    elif output_.shape[1] >= 2:  # vector output ##############################

        for i, c in enumerate(components):
            c_output = output_.extract([c])
            if i == 0:
                gradients = grad_scalar_output(c_output, input_, d)
            else:
                gradients = gradients.append(
                    grad_scalar_output(c_output, input_, d))
    else:
        raise NotImplementedError

    return gradients


def div(output_, input_, components=None, d=None):
    """
    Perform divergence operation. The operator works for vectorial functions,
    with multiple input coordinates.

    :param LabelTensor output_: the output tensor onto which computing the
        divergence.
    :param LabelTensor input_: the input tensor with respect to which computing
        the divergence.
    :param list(str) components: the name of the output variables to calculate
        the divergence for. It should be a subset of the output labels. If None,
        all the output variables are considered. Default is None.
    :param list(str) d: the name of the input variables on which the divergence
        is calculated. d should be a subset of the input labels. If None, all
        the input variables are considered. Default is None.

    :raises TypeError: div operator works only for LabelTensor.
    :raises ValueError: div operator works only for vector fields.
    :raises ValueError: div operator must derive all components with
        respect to all coordinates.
    :return: the divergenge tensor.
    :rtype: LabelTensor
    """
    if not isinstance(input_, LabelTensor):
        raise TypeError

    if d is None:
        d = input_.labels

    if components is None:
        components = output_.labels

    if output_.shape[1] < 2 or len(components) < 2:
        raise ValueError('div supported only for vector fields')

    if len(components) != len(d):
        raise ValueError

    grad_output = grad(output_, input_, components, d)
    div = torch.zeros(input_.shape[0], 1, device=output_.device)
    labels = [None] * len(components)
    for i, (c, d) in enumerate(zip(components, d)):
        c_fields = f'd{c}d{d}'
        div[:, 0] += grad_output.extract(c_fields).sum(axis=1)
        labels[i] = c_fields

    div = div.as_subclass(LabelTensor)
    div.labels = ['+'.join(labels)]
    return div


def laplacian(output_, input_, components=None, d=None, method='std'):
    """
    Compute Laplace operator. The operator works for vectorial and
    scalar functions, with multiple input coordinates.

    :param LabelTensor output_: the output tensor onto which computing the
        Laplacian.
    :param LabelTensor input_: the input tensor with respect to which computing
        the Laplacian.
    :param list(str) components: the name of the output variables to calculate
        the Laplacian for. It should be a subset of the output labels. If None,
        all the output variables are considered. Default is None.
    :param list(str) d: the name of the input variables on which the Laplacian
        is calculated. d should be a subset of the input labels. If None, all
        the input variables are considered. Default is None.
    :param str method: used method to calculate Laplacian, defaults to 'std'.
    :raises ValueError: for vectorial field derivative with respect to
        all coordinates must be performed.
    :raises NotImplementedError: 'divgrad' not implemented as method.
    :return: The tensor containing the result of the Laplacian operator.
    :rtype: LabelTensor
    """
    if d is None:
        d = input_.labels

    if components is None:
        components = output_.labels

    if len(components) != len(d) and len(components) != 1:
        raise ValueError

    if method == 'divgrad':
        raise NotImplementedError('divgrad not implemented as method')
        # TODO fix
        # grad_output = grad(output_, input_, components, d)
        # result = div(grad_output, input_, d=d)
    elif method == 'std':

        if len(components) == 1:
            grad_output = grad(output_, input_, components=components, d=d)
            result = torch.zeros(output_.shape[0], 1, device=output_.device)
            for i, label in enumerate(grad_output.labels):
                gg = grad(grad_output, input_, d=d, components=[label])
                result[:, 0] += super(torch.Tensor,
                                      gg.T).__getitem__(i)  # TODO improve
            labels = [f'dd{components[0]}']

        else:
            result = torch.empty(input_.shape[0],
                                 len(components),
                                 device=output_.device)
            labels = [None] * len(components)
            for idx, (ci, di) in enumerate(zip(components, d)):

                if not isinstance(ci, list):
                    ci = [ci]
                if not isinstance(di, list):
                    di = [di]

                grad_output = grad(output_, input_, components=ci, d=di)
                result[:, idx] = grad(grad_output, input_, d=di).flatten()
                labels[idx] = f'dd{ci}dd{di}'

    result = result.as_subclass(LabelTensor)
    result.labels = labels
    return result


def advection(output_, input_, velocity_field, components=None, d=None):
    """
    Perform advection operation. The operator works for vectorial functions,
    with multiple input coordinates.

    :param LabelTensor output_: the output tensor onto which computing the
        advection.
    :param LabelTensor input_: the input tensor with respect to which computing
        the advection.
    :param str velocity_field: the name of the output variables which is used
        as velocity field. It should be a subset of the output labels.
    :param list(str) components: the name of the output variables to calculate
        the Laplacian for. It should be a subset of the output labels. If None,
        all the output variables are considered. Default is None.
    :param list(str) d: the name of the input variables on which the advection
        is calculated. d should be a subset of the input labels. If None, all
        the input variables are considered. Default is None.
    :return: the tensor containing the result of the advection operator.
    :rtype: LabelTensor
    """
    if d is None:
        d = input_.labels

    if components is None:
        components = output_.labels

    tmp = grad(output_, input_, components, d).reshape(-1, len(components),
                                                       len(d)).transpose(0, 1)

    tmp *= output_.extract(velocity_field)
    return tmp.sum(dim=2).T



"""Module for the ParametricProblem class"""
from abc import abstractmethod

from .abstract_problem import AbstractProblem


class ParametricProblem(AbstractProblem):
    """
    The class for the definition of parametric problems, i.e., problems
    with parameters among the input variables.

    Here's an example of a spatial parametric ODE problem, i.e., a spatial
    ODE problem with an additional parameter `alpha` as coefficient of the
    derivative term.

    :Example:
        >>> from pina.problem import SpatialProblem, ParametricProblem
        >>> from pina.operators import grad
        >>> from pina.equations import Equation, FixedValue
        >>> from pina import Condition
        >>> from pina.geometry import CartesianDomain
        >>> import torch
        >>>
        >>>
        >>> class ParametricODE(SpatialProblem, ParametricProblem):
        >>>
        >>>     output_variables = ['u']
        >>>     spatial_domain = CartesianDomain({'x': [0, 1]})
        >>>     parameter_domain = CartesianDomain({'alpha': [1, 10]})
        >>>
        >>>     def ode_equation(input_, output_):
        >>>         u_x = grad(output_, input_, components=['u'], d=['x'])
        >>>         u = output_.extract(['u'])
        >>>         alpha = input_.extract(['alpha'])
        >>>         return alpha * u_x - u
        >>>
        >>>     conditions = {
        >>>         'x0': Condition(CartesianDomain({'x': 0, 'alpha':[1, 10]}), FixedValue(1.)),
        >>>         'D': Condition(CartesianDomain({'x': [0, 1], 'alpha':[1, 10]}), Equation(ode_equation))}
    """

    @abstractmethod
    def parameter_domain(self):
        """
        The parameters' domain of the problem.
        """
        pass

    @property
    def parameters(self):
        """
        The parameters' variables of the problem.
        """
        return self.parameter_domain.variables



import torch


def check_point(x, current_stride, dim):
    max_stride = current_stride + dim
    indeces = torch.logical_and(x[..., :-1] < max_stride, x[..., :-1]
                                >= current_stride).all(dim=-1)
    return indeces


def map_points_(x, filter_position):
    """Mapping function n dimensional case

    :param x: input data of two dimension
    :type x: torch.tensor
    :param filter_position: position of the filter
    :type dim: list[numeric]
    :return: data mapped inplace
    :rtype: torch.tensor
    """
    x.add_(-filter_position)

    return x


def optimizing(f):
    """Decorator for calling a function just once

    :param f: python function
    :type f: function
    """

    def wrapper(*args, **kwargs):

        if kwargs['type'] == 'forward':
            if not wrapper.has_run_inverse:
                wrapper.has_run_inverse = True
                return f(*args, **kwargs)

        if kwargs['type'] == 'inverse':
            if not wrapper.has_run:
                wrapper.has_run = True
                return f(*args, **kwargs)

    wrapper.has_run_inverse = False
    wrapper.has_run = False

    return wrapper



import torch
from .location import Location
from pina.geometry import CartesianDomain
from pina import LabelTensor
from ..utils import check_consistency


class SimplexDomain(Location):
    """PINA implementation of a Simplex."""

    def __init__(self, simplex_matrix, sample_surface=False):
        """
        :param simplex_matrix: A matrix of LabelTensor objects representing
            a vertex of the simplex (a tensor), and the coordinates of the
            point (a list of labels).

        :type simplex_matrix: list[LabelTensor]
        :param sample_surface: A variable for choosing sample strategies. If
            ``sample_surface=True`` only samples on the Simplex surface
            frontier are taken. If ``sample_surface=False``, no such criteria
            is followed.

        :type sample_surface: bool

        .. warning::
            Sampling for dimensions greater or equal to 10 could result
            in a shrinking of the simplex, which degrades the quality
            of the samples. For dimensions higher than 10, other algorithms
            for sampling should be used.

        :Example:
            >>> spatial_domain = SimplexDomain(
                    [
                        LabelTensor(torch.tensor([[0, 0]]), labels=["x", "y"]),
                        LabelTensor(torch.tensor([[1, 1]]), labels=["x", "y"]),
                        LabelTensor(torch.tensor([[0, 2]]), labels=["x", "y"]),
                    ], sample_surface = True
                )
        """

        # check consistency of sample_surface as bool
        check_consistency(sample_surface, bool)
        self._sample_surface = sample_surface

        # check consistency of simplex_matrix as list or tuple
        check_consistency([simplex_matrix], (list, tuple))

        # check everything within simplex_matrix is a LabelTensor
        check_consistency(simplex_matrix, LabelTensor)

        # check consistency of labels
        matrix_labels = simplex_matrix[0].labels
        if not all(vertex.labels == matrix_labels for vertex in simplex_matrix):
            raise ValueError(f"Labels don't match.")

        # check consistency dimensions
        dim_simplex = len(matrix_labels)
        if len(simplex_matrix) != dim_simplex + 1:
            raise ValueError(
                "An n-dimensional simplex is composed by n + 1 tensors of dimension n."
            )

        # creating vertices matrix
        self._vertices_matrix = LabelTensor.vstack(simplex_matrix)

        # creating basis vectors for simplex
        # self._vectors_shifted = (
        #     (self._vertices_matrix.T)[:, :-1] - (self._vertices_matrix.T)[:, None, -1]
        # )  ### TODO: Remove after checking

        vert = self._vertices_matrix
        self._vectors_shifted = (vert[:-1] - vert[-1]).T

        # build cartesian_bound
        self._cartesian_bound = self._build_cartesian(self._vertices_matrix)

    @property
    def variables(self):
        return self._vertices_matrix.labels

    def _build_cartesian(self, vertices):
        """
        Build Cartesian border for Simplex domain to be used in sampling.
        :param vertex_matrix: matrix of vertices
        :type vertices: list[list]
        :return: Cartesian border for triangular domain
        :rtype: CartesianDomain
        """

        span_dict = {}

        for i, coord in enumerate(self.variables):
            sorted_vertices = sorted(vertices, key=lambda vertex: vertex[i])
            # respective coord bounded by the lowest and highest values
            span_dict[coord] = [
                float(sorted_vertices[0][i]),
                float(sorted_vertices[-1][i])
            ]

        return CartesianDomain(span_dict)

    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the simplex.
        Uses the algorithm described involving barycentric coordinates:
        https://en.wikipedia.org/wiki/Barycentric_coordinate_system.

        :param point: Point to be checked.
        :type point: LabelTensor
        :param check_border: Check if the point is also on the frontier
            of the simplex, default ``False``.
        :type check_border: bool
        :return: Returning ``True`` if the point is inside, ``False`` otherwise.
        :rtype: bool

        .. note::
            When ``sample_surface`` in the ``__init()__``
            is set to ``True``, then the method only checks
            points on the surface, and not inside the domain.
        """

        if not all(label in self.variables for label in point.labels):
            raise ValueError("Point labels different from constructor"
                             f" dictionary labels. Got {point.labels},"
                             f" expected {self.variables}.")

        point_shift = point - self._vertices_matrix[-1]
        point_shift = point_shift.tensor.reshape(-1, 1)

        # compute barycentric coordinates
        lambda_ = torch.linalg.solve(self._vectors_shifted * 1.0,
                                     point_shift * 1.0)
        lambda_1 = 1.0 - torch.sum(lambda_)
        lambdas = torch.vstack([lambda_, lambda_1])

        # perform checks
        if not check_border:
            return all(torch.gt(lambdas, 0.0)) and all(torch.lt(lambdas, 1.0))

        return all(torch.ge(lambdas, 0)) and (any(torch.eq(lambdas, 0))
                                              or any(torch.eq(lambdas, 1)))

    def _sample_interior_randomly(self, n, variables):
        """
        Randomly sample points inside a simplex of arbitrary
        dimension, without the boundary.
        :param int n: Number of points to sample in the shape.
        :param variables: pinn variable to be sampled, defaults to ``all``.
        :type variables: str or list[str], optional
        :return: Returns tensor of n sampled points.
        :rtype: torch.Tensor
        """

        # =============== For Developers ================ #
        #
        # The sampling startegy used is fairly simple.
        # First we sample a random vector from the hypercube
        # which contains the simplex. Then, if the point
        # sampled is inside the simplex, we add it as a valid
        # one.
        #
        # =============================================== #

        sampled_points = []
        while len(sampled_points) < n:
            sampled_point = self._cartesian_bound.sample(n=1,
                                                         mode="random",
                                                         variables=variables)

            if self.is_inside(sampled_point, self._sample_surface):
                sampled_points.append(sampled_point)
        return torch.cat(sampled_points, dim=0)

    def _sample_boundary_randomly(self, n):
        """
        Randomly sample points on the boundary of a simplex
        of arbitrary dimensions.
        :param int n: Number of points to sample in the shape.
        :return: Returns tensor of n sampled points
        :rtype: torch.Tensor
        """

        # =============== For Developers ================ #
        #
        # The sampling startegy used is fairly simple.
        # We first sample the lambdas in [0, 1] domain,
        # we then set to zero only one lambda, and normalize.
        # Finally, we compute the matrix product between the
        # lamdas and the vertices matrix.
        #
        # =============================================== #

        sampled_points = []

        while len(sampled_points) < n:
            # extract number of vertices
            number_of_vertices = self._vertices_matrix.shape[0]
            # extract idx lambda to set to zero randomly
            idx_lambda = torch.randint(low=0,
                                       high=number_of_vertices,
                                       size=(1, ))
            # build lambda vector
            # 1. sampling [1, 2)
            lambdas = torch.rand((number_of_vertices, 1))
            # 2. setting lambdas[idx_lambda] to 0
            lambdas[idx_lambda] = 0
            # 3. normalize
            lambdas /= lambdas.sum()
            # 4. compute dot product
            sampled_points.append(self._vertices_matrix.T @ lambdas)
        return torch.cat(sampled_points, dim=1).T

    def sample(self, n, mode="random", variables="all"):
        """
        Sample n points from Simplex domain.

        :param int n: Number of points to sample in the shape.
        :param str mode: Mode for sampling, defaults to ``random``. Available modes include: ``random``.
        :param variables: Variables to be sampled, defaults to ``all``.
        :type variables: str | list[str]
        :return: Returns ``LabelTensor`` of n sampled points.
        :rtype: LabelTensor

        .. warning::
            When ``sample_surface = True`` in the initialization, all
            the variables are sampled, despite passing different once
            in ``variables``.
        """

        if mode in ["random"]:
            if self._sample_surface:
                sample_pts = self._sample_boundary_randomly(n)
            else:
                sample_pts = self._sample_interior_randomly(n, variables)

        else:
            raise NotImplementedError(f"mode={mode} is not implemented.")

        return LabelTensor(sample_pts, labels=self.variables)


""" Module for Loss class """

from abc import ABCMeta, abstractmethod
from torch.nn.modules.loss import _Loss
import torch
from .utils import check_consistency

__all__ = ['LossInterface', 'LpLoss', 'PowerLoss']


class LossInterface(_Loss, metaclass=ABCMeta):
    """
    The abstract ``LossInterface`` class. All the class defining a PINA Loss
    should be inheritied from this class.
    """

    def __init__(self, reduction='mean'):
        """
        :param str reduction: Specifies the reduction to apply to the output:
            ``none`` | ``mean`` | ``sum``. When ``none``: no reduction 
            will be applied, ``mean``: the sum of the output will be divided
            by the number of elements in the output, ``sum``: the output will
            be summed. Note: ``size_average`` and ``reduce`` are in the
            process of being deprecated, and in the meantime, specifying either of
            those two args will override ``reduction``. Default: ``mean``.
        """
        super().__init__(reduction=reduction, size_average=None, reduce=None)

    @abstractmethod
    def forward(self, input, target):
        """Forward method for loss function.

        :param torch.Tensor input: Input tensor from real data.
        :param torch.Tensor target: Model tensor output.
        :return: Loss evaluation.
        :rtype: torch.Tensor
        """
        pass

    def _reduction(self, loss):
        """Simple helper function to check reduction

        :param reduction: Specifies the reduction to apply to the output:
            ``none`` | ``mean`` | ``sum``. When ``none``: no reduction 
            will be applied, ``mean``: the sum of the output will be divided
            by the number of elements in the output, ``sum``: the output will
            be summed. Note: ``size_average`` and ``reduce`` are in the
            process of being deprecated, and in the meantime, specifying either of
            those two args will override ``reduction``. Default: ``mean``.
        :type reduction: str
        :param loss: Loss tensor for each element.
        :type loss: torch.Tensor
        :return: Reduced loss.
        :rtype: torch.Tensor
        """
        if self.reduction == "none":
            ret = loss
        elif self.reduction == "mean":
            ret = torch.mean(loss, keepdim=True, dim=-1)
        elif self.reduction == "sum":
            ret = torch.sum(loss, keepdim=True, dim=-1)
        else:
            raise ValueError(self.reduction + " is not valid")
        return ret


class LpLoss(LossInterface):
    r"""
    The Lp loss implementation class. Creates a criterion that measures
    the Lp error between each element in the input :math:`x` and
    target :math:`y`.

    The unreduced (i.e. with ``reduction`` set to ``none``) loss can
    be described as:

    .. math::
        \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
        l_n = \left[\sum_{i=1}^{D} \left| x_n^i - y_n^i \right|^p \right],
    
    If ``'relative'`` is set to true:

    .. math::
        \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
        l_n = \frac{ [\sum_{i=1}^{D} | x_n^i - y_n^i|^p] }{[\sum_{i=1}^{D}|y_n^i|^p]},

    where :math:`N` is the batch size. If ``reduction`` is not ``none``
    (default ``mean``), then:

    .. math::
        \ell(x, y) =
        \begin{cases}
            \operatorname{mean}(L), & \text{if reduction} = \text{`mean';}\\
            \operatorname{sum}(L),  & \text{if reduction} = \text{`sum'.}
        \end{cases}

    :math:`x` and :math:`y` are tensors of arbitrary shapes with a total
    of :math:`n` elements each.

    The sum operation still operates over all the elements, and divides by :math:`n`.

    The division by :math:`n` can be avoided if one sets ``reduction`` to ``sum``.
    """

    def __init__(self, p=2, reduction='mean', relative=False):
        """
        :param int p: Degree of Lp norm. It specifies the type of norm to
            be calculated. See `list of possible orders in torch linalg
            <https://pytorch.org/docs/stable/generated/torch.linalg.norm.html#torch.linalg.norm>`_ to
            for possible degrees. Default 2 (euclidean norm).
        :param str reduction: Specifies the reduction to apply to the output:
            ``none`` | ``mean`` | ``sum``. ``none``: no reduction 
            will be applied, ``mean``: the sum of the output will be divided
            by the number of elements in the output, ``sum``: the output will
            be summed.
        :param bool relative: Specifies if relative error should be computed.
        """
        super().__init__(reduction=reduction)

        # check consistency
        check_consistency(p, (str,int,float))
        check_consistency(relative, bool)

        self.p = p
        self.relative = relative

    def forward(self, input, target):
        """Forward method for loss function.

        :param torch.Tensor input: Input tensor from real data.
        :param torch.Tensor target: Model tensor output.
        :return: Loss evaluation.
        :rtype: torch.Tensor
        """
        loss = torch.linalg.norm((input - target), ord=self.p, dim=-1)
        if self.relative:
            loss = loss / torch.linalg.norm(input, ord=self.p, dim=-1)
        return self._reduction(loss)


class PowerLoss(LossInterface):
    r"""
    The PowerLoss loss implementation class. Creates a criterion that measures
    the error between each element in the input :math:`x` and
    target :math:`y` powered to a specific integer.

    The unreduced (i.e. with ``reduction`` set to ``none``) loss can
    be described as:

    .. math::
        \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
        l_n = \frac{1}{D}\left[\sum_{i=1}^{D} \left| x_n^i - y_n^i \right|^p \right],
    
    If ``'relative'`` is set to true:

    .. math::
        \ell(x, y) = L = \{l_1,\dots,l_N\}^\top, \quad
        l_n = \frac{ \sum_{i=1}^{D} | x_n^i - y_n^i|^p }{\sum_{i=1}^{D}|y_n^i|^p},

    where :math:`N` is the batch size. If ``reduction`` is not ``none``
    (default ``mean``), then:

    .. math::
        \ell(x, y) =
        \begin{cases}
            \operatorname{mean}(L), & \text{if reduction} = \text{`mean';}\\
            \operatorname{sum}(L),  & \text{if reduction} = \text{`sum'.}
        \end{cases}

    :math:`x` and :math:`y` are tensors of arbitrary shapes with a total
    of :math:`n` elements each.

    The sum operation still operates over all the elements, and divides by :math:`n`.

    The division by :math:`n` can be avoided if one sets ``reduction`` to ``sum``.
    """

    def __init__(self, p=2, reduction='mean', relative=False):
        """
        :param int p: Degree of Lp norm. It specifies the type of norm to
            be calculated. See `list of possible orders in torch linalg
            <https://pytorch.org/docs/stable/generated/torch.linalg.norm.html#torch.linalg.norm>`_ to
            see the possible degrees. Default 2 (euclidean norm).
        :param str reduction: Specifies the reduction to apply to the output:
            ``none`` | ``mean`` | ``sum``. When ``none``: no reduction 
            will be applied, ``mean``: the sum of the output will be divided
            by the number of elements in the output, ``sum``: the output will
            be summed.
        :param bool relative: Specifies if relative error should be computed.
        """
        super().__init__(reduction=reduction)

        # check consistency
        check_consistency(p, (str, int, float))
        self.p = p
        check_consistency(relative, bool)
        self.relative = relative

    def forward(self, input, target):
        """Forward method for loss function.

        :param torch.Tensor input: Input tensor from real data.
        :param torch.Tensor target: Model tensor output.
        :return: Loss evaluation.
        :rtype: torch.Tensor
        """
        loss = torch.abs((input - target)).pow(self.p).mean(-1)
        if self.relative:
            loss = loss / torch.abs(input).pow(self.p).mean(-1)
        return self._reduction(loss)
